Require Import BakerLemma.
Require Import Baker.
Require Import BakerAlloc.
Require Import BakerBarriers.
Require Import VGC.


Set Implicit Arguments.


Module BakerGCImpl <: VIEW_GC with Module I := BMGI.

  Module I := BMGI.

  Definition repr := repr.


  Lemma atomRegEq : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    In r rr ->
    I.atomP (RFile.get rf r) \/ I.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.
  Proof.
    intros h vs rf rr A r Hp I D.
    unfold repr, Baker.repr in Hp; lle Hp; clear Hp.
    destruct H3 as [_ [_ [_ [_ [_ [Req _]]]]]].
    destruct (Req _ I) as [X _]; auto.
  Qed.


  Lemma setNonRootOk : forall h vs rf rr r v,
    h ||- repr vs rf rr ->
    ~In r (rgcInfo::rr) ->
    h ||- repr vs (RFile.set rf r v) rr.
  Proof.
    unfold repr, Baker.repr.
    intros h vs rf rr r v Hp I; lle Hp.
    rename x1 into toSt; rename x3 into scan; rename x4 into free; rename x5 into alloc;
      rename x2 into toEnd; rename x6 into phi; rename x7 into W; rename x8 into M; 
	rename x into frSt; rename x0 into frEnd; rename x9 into vh.
    destruct H1 as [NI [als [alFr [WMeq [Iso [Req VsEq]]]]]].
    existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl scan; 
      existsl free; existsl alloc; existsl phi; existsl W; existsl M; existsl vh.
    replace (RFile.get (RFile.set rf r v))
      with (rget (rset (h, rf) r v)) by auto.
    lli.
    repeat (apply conj; [auto | ]).
    intros r' IRR.
    destruct (regCompare r r').
    elf; subst r'; apply I; right; assumption.
    stSimpl.
    hyr Req (Req _ IRR).
    replace (rget (h, rf) r') with (RFile.get rf r'); [assumption | reflexivity].
    assumption.
    scm.
    destruct (regCompare r rgcInfo).
    elf; subst r; apply I; left; reflexivity.
    stSimpl; assumption.
  Qed.


  Lemma rootCopyOk : forall h vs rf rr rs rd,
    h ||- repr vs rf rr ->
    In rs rr -> 
    ~In rd (zero::rgcInfo::nil) ->
    h ||- repr (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).
  Proof.
    unfold repr, Baker.repr.
    intros h vs rf rr rs rd Hp I0 Nz.
    assert (NE1 : rd <> zero) by (intro E; subst rd; intuition).
    assert (NE2 : rd <> rgcInfo) by (intro E; subst rd; intuition).
    lle Hp.
    rename x1 into toSt; rename x3 into scan; rename x4 into free; rename x5 into alloc;
      rename x2 into toEnd; rename x6 into phi; rename x7 into W; rename x8 into M; 
	rename x into frSt; rename x0 into frEnd; rename x9 into vh.
    destruct H1 as [NI [als [alFr [WMeq [Iso [Req VsEq]]]]]].
    existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl scan; 
      existsl free; existsl alloc; existsl phi; existsl W; existsl M; existsl vh.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)))
      with (rget (rset (h, rf) rd (rget (h, rf) rs))) by reflexivity.
    lli.
    constructor.
    intro I; simpl in I; destruct I as [I | I].
    apply Nz; right; left; auto.
    auto.
    repeat (constructor; [assumption | ]).
    constructor.
    intros r II; simpl in II; destruct II.
    subst r; stSimpl; apply (Req _ I0).
    destruct (regCompare r rd).
    subst r; stSimpl; apply (Req _ I0).
    stSimpl; apply (Req _ H). assumption.
    stSimpl; assumption.
  Qed.


  Lemma rootWeaken : forall h vs rf rr rr',
    h ||- repr vs rf rr ->
    incl rr' rr ->
    h ||- repr vs rf rr'.
  Proof.
    unfold repr, Baker.repr.
    intros h vs rf rr rr' Hp IR; lle Hp.
    rename x1 into toSt; rename x3 into scan; rename x4 into free; rename x5 into alloc;
      rename x2 into toEnd; rename x6 into phi; rename x7 into W; rename x8 into M; 
	rename x into frSt; rename x0 into frEnd; rename x9 into vh.
    destruct H1 as [NI [als [alFr [WMeq [Iso [Req VsEq]]]]]].
    existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl scan; 
      existsl free; existsl alloc; existsl phi; existsl W; existsl M; existsl vh.
    lli.
    constructor.
    intro I; apply NI; apply IR; assumption.
    repeat (constructor; [assumption | ]).
    constructor.
    intros r IRR; apply (Req r); apply IR; assumption.
    assumption.
    assumption.
  Qed.


  Lemma addAtomRoot : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    ~In r (zero::I.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- repr (rset vs r (RFile.get rf r)) rf (r::rr) ** A.
  Proof.
    unfold repr; unfold Baker.repr.
    intros h vs rf rr A r Hp NI Atom.
    assert (NE1 : r <> zero) by (intro E; subst r; intuition).
    assert (NE2 : r <> rgcInfo) by (intro E; subst r; intuition).
    lle Hp.
    rename x1 into toSt; rename x3 into scan; rename x4 into free; rename x5 into alloc;
      rename x2 into toEnd; rename x6 into phi; rename x7 into W; rename x8 into M; 
	rename x into frSt; rename x0 into frEnd; rename x9 into vh.
    destruct H3 as [NI' [als [alFr [WMeq [Iso [Req VsEq]]]]]].
    hyr Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR
      (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp))))))))); scm; lle Hp; clear A.
    existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl scan; 
      existsl free; existsl alloc; existsl phi; existsl W; existsl M; existsl vh.    
    lli; [clear Hp | assumption].
    repeat (apply conj; [auto | ]).
    simpl; intros [I | I]; [subst r; apply NI; right; left; reflexivity | apply NI'; assumption].
    intros r0 I.
    destruct (regCompare r0 r) as [E | E].
    destruct (regCompare r0 zero) as [E' | E'].
    subst r0 r; elf; apply NI; left; reflexivity.
    subst r0; stSimpl; auto.
    constructor; [reflexivity | left; assumption].
    destruct I as [I | I]; [elf; auto | stSimpl; apply (Req r0); assumption].
    auto.
  Qed.


  Lemma reprIsBakerStOk : forall s vs A,
    s |||- repr vs (rfileOf s) (rroot :: nil) ** A ->
    bakerStOk vs A s.
  Proof.
    intros s vs A Hp.
    unfold repr, Baker.repr in Hp;
      replace (RFile.get (rfileOf s)) with (rget s) in Hp by reflexivity.
    assert (x := 0).
    lle Hp.
    rename x2 into toSt; rename x4 into scan; rename x5 into free; rename x6 into alloc;
      rename x3 into toEnd; rename x7 into phi; rename x8 into W; rename x9 into M; 
	rename x0 into frSt; rename x1 into frEnd; rename x10 into vh.
    clear x.
    exists frSt; exists frEnd; exists toSt; exists toEnd; exists scan; exists free; 
      exists alloc.
    destruct H3 as [_ [AlS [Al [WMeq [Iso [rootOk Veq]]]]]].
    destruct (rootOk rroot).
    simpl; auto.
    constructor.
    assumption.
    exists phi; exists W; exists M.
    intuition.
    intuition.

    (* show the heap is okay *)
    destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp'].
    sconjLift2 Hp'; lle Hp'; lli; exact Hp'.
    
    (* show the virtual heap is okay *)
    assert (Vhp := heapEqvImplR Hp).

    assert (Sub : Heap.subHeap vh (heapOf s)).
      destruct (aconjElim (sconjDistrib2 Hp)) as [HpX _]; clear Hp.
      eapply sconjEqSubHeap; eauto.
    
    assert (Meven : forall x, set_In x M -> even x).
      clear rootOk Veq H Vhp.
      assert (Hp' : s |||- (foralls x in M, x |->__) ** top).
        sconjLift2 Hp; scm; exact (mapHpIsObjs Hp).
      clear Hp Iso H0.
      intros x Ix; apply ptrPeven.
      assert (Hp : s |||- (x |->__ ** top) ** top).
      scm; hyr Hp (fasElim Ix Hp); scm.
      lle Hp; apply (contElim1L Hp).

    set (B := rangeObjs toSt scan) in * .
    set (G := rangeObjs scan free) in * .
    set (B' := rangeObjs alloc toEnd) in * .
    set (mphi := fun x : POnat.OneTypeNat.B =>
      match phi x with Some y => y | None => x end) in * .

    assert (disj : set_disj (B |_| G |_| B') M).
      destruct (aconjElim (sconjDistrib2 Hp)) as [_ HpX]; clear Hp Vhp.
      assert (HpY : s |||- fasd (B |_| G |_| B') ** fasd M ** top).
        lle HpX; sconjLift3 HpX; sconjLift3 HpX; hyr HpX (sconjAssocR HpX);
          sconjLift3 HpX; sconjLift2 HpX; scm.
        assert (Hp' : h ||- objHp ((W |_| M) |_| B |_| G |_| B') (B |_| G |_| B')).
          set (U :=  (W |_| M) |_| B |_| G |_| B') in * ;
            unfold hpredTrue; rewrite (set_eqvUnionSymm G B'); 
	      rewrite <- (set_eqvUnionAssoc B B' G); apply objHpMerge; scm.
          apply (objHpWeaken Hp); clear Hp; subst U; intros x Ix; apply set_union_intro2; auto.
        unfold objHp in Hp'; llMatch; hpunfoldIn Hp0; lleim.
        hyr Hp (mapHpIsObjs Hp); llMatch; lleim.
      apply (fasFasDisj HpY); auto.

    clear AlS Al WMeq rootOk H H0 Hp frSt frEnd.
    eapply heapEqvTrue. eauto.
    clear vs Veq.
    sconjLift3of3 Vhp; sconjLift2 Vhp; hyr Vhp (sconjAssocR Vhp); sconjLift2of2 Vhp.
    apply (sconjMapRangeImplS Vhp); clear Vhp; intros h' Sub' Hp.
    apply (fasMapRangeImplIS Hp); clear Hp.
    intros h x Sub'' _ Hp.
    assert (Sub0 : Heap.subHeap h (heapOf s)).
    eapply Heap.subHeapTrans; eauto.
    eapply Heap.subHeapTrans; eauto.
    clear Sub Sub' Sub'' h'.
    hpunfoldIn Hp; lle Hp; assert (Hp' := subHeapImpl Hp Sub0); lle Hp'.
    unfold objCopiedH; unfold fieldCopiedH; lli.
    scm.
    instantiate (1 := x1).
    2:sconjLift2 Hp'; scm.
    apply (sconjMapRangeImpl Hp); intros h' Hp''.
    (* first field *)
    replace x0 with (mphi x0).
    apply (contMapRangeImpl Hp'').
    subst mphi; cbv beta.
    destruct Iso as [Iso _]; destruct (Iso x0) as [_ Iso']; clear Iso.
    destruct (phi x0).
    assert (IM : set_In x0 M).
    apply Iso'; discriminate.
    clear Iso'.
    destruct H3.
    elf; apply (not_even_and_odd x0); eauto.
    elf; apply (set_disj_elim disj x0); assumption.
    reflexivity.

    (* second field *)
    replace x1 with (mphi x1).
    apply (contMapRangeImpl Hp'').
    subst mphi; cbv beta.
    destruct Iso as [Iso _]; destruct (Iso x1) as [_ Iso']; clear Iso.
    destruct (phi x1).
    assert (IM : set_In x1 M).
    apply Iso'; discriminate.
    clear Iso'.
    destruct H4.
    elf; apply (not_even_and_odd x1); eauto.
    elf; apply (set_disj_elim disj x1); assumption.
    reflexivity.

    (* show G and W respect the map *)
    assert (Hp' : h' ||- objHp ((W |_| M) |_| B |_| G |_| B') (G |_| W)).
    apply objHpMerge; apply sconjComm; scm.
    apply (objHpWeaken Hp); intros x Ix; apply set_union_intro1; assumption.
    clear Hp; rename Hp' into Hp; unfold objHp in Hp.
    apply (fasMapRangeImplIS Hp); clear Hp; intros h x Sub'' _ Hp.
    hpunfold; hpunfoldIn Hp; apply (sconjMapRangeImplS Hp); clear Hp; intros h'' Sub3 Hp.

    (* field 1 *)
    lle Hp; lli.
    2:instantiate (1 := mphi x0); apply (contMapRangeImpl Hp).
    assert (Sub0 : Heap.subHeap h'' (heapOf s)).
    eapply Heap.subHeapTrans; eauto.
    eapply Heap.subHeapTrans; eauto.
    eapply Heap.subHeapTrans; eauto.
    clear Sub Sub' Sub'' Sub3 h h'.
    exists x0; constructor.
    eapply subHeapImpl; eauto.
    unfold mapX.
    assert (IO := isOddOk x0); destruct (isOdd x0).
    subst mphi; simpl.
    destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
    elf; apply (not_even_and_odd x0); auto; apply Meven; apply Q; discriminate.
    reflexivity.
    subst mphi; unfold map_union; destruct (phi x0); reflexivity.

    (* second field *)
    lle Hp; lli.
    2:instantiate (1 := mphi x0); apply (contMapRangeImpl Hp).
    assert (Sub0 : Heap.subHeap h'' (heapOf s)).
    eapply Heap.subHeapTrans; eauto.
    eapply Heap.subHeapTrans; eauto.
    eapply Heap.subHeapTrans; eauto.
    clear Sub Sub' Sub'' Sub3 h h'.
    exists x0; constructor.
    eapply subHeapImpl; eauto.
    unfold mapX.
    assert (IO := isOddOk x0); destruct (isOdd x0).
    subst mphi; simpl.
    destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
    elf; apply (not_even_and_odd x0); auto; apply Meven; apply Q; discriminate.
    reflexivity.
    subst mphi; unfold map_union; destruct (phi x0); reflexivity.

  Qed.


  Lemma bakerStOkIsRepr : forall s vs A,
    bakerStOk vs A s ->
    s |||- repr vs (rfileOf s) (rroot :: nil) ** A.
  Proof.
    unfold bakerStOk.
    intros s vs A [frStart [frEnd [toStart [toEnd [scan [free [alloc X]]]]]]].
    destruct X as [rootEq Bok].
    destruct Bok as [phi [W [M [Al1 [Al2 [Wmeq [Iso [rootOk [HpOk Vok]]]]]]]]].
    sconjLift7of7 HpOk; apply sconjComm; apply (sconjImplS HpOk); clear HpOk.
    auto.
    intros h Sub Hp; unfold repr; unfold Baker.repr.
    existsl frStart; existsl frEnd; existsl toStart; existsl toEnd;
    existsl scan; existsl free; existsl alloc; 
    existsl phi; existsl W; existsl M.
    set (B := rangeObjs toStart scan) in * ;
      set (G := rangeObjs scan free) in * ;
	set (B' := rangeObjs alloc toEnd) in * ;
	  set (F := rangeObjs free alloc) in * .
    sconjLift4 Hp; sconjLift4 Hp; sconjLift4 Hp; hyr Hp (sconjAssocR (sconjAssocR Hp)).
    destruct (sconjExtract Hp) as [h' [Hp' HpH]]; clear Hp; rename Hp' into Hp;
      lle HpH.
    existsl h'.
    replace (RFile.get (rfileOf s)) with (rget s); [idtac | reflexivity].
    
    assert (HpD : h ||- (fasd W ** fasd (B |_| B') ** fasd G) ** fasd M ** fasd F ** top).
    assert (HpH' : h' ||- fasd W ** fasd (B |_| B') ** fasd G).
    scm; hyr Hp0 (objHpIsObjs Hp0); llMatch; lleim.
    hyr Hp (sconjInsert Hp HpH'); scm; [hyr Hp (mapHpIsObjs Hp) | idtac]; llMatch; lleim.

    assert (BBGeqv : set_eqv (B |_| G |_| B') ((B |_| B') |_| G)).
    rewrite (set_eqvUnionSymm G B'); rewrite <- (set_eqvUnionAssoc B B' G); 
      apply set_eqvRefl.

    assert (PeqB : forall z, okFieldVal (B |_| G |_| B') z -> phi z = None).
    intros z okZ; destruct Iso as [Iso _]; destruct (Iso z) as [_ Q];
      destruct (phi z).
    2:reflexivity.
    assert (Q' : set_In z M); [apply Q; discriminate | clear Q; elf].
    destruct okZ.
    apply (not_even_and_odd z).
    apply ptrPeven; assert (Hp' : h ||- (z |->__ ** top) ** top).
    sconjLift2 Hp; scm; hyr Hp (fasElim Q' (mapHpIsObjs Hp)); lleim.
    lle Hp'; apply (contElim1L Hp').
    assumption.

    assert (MBGBdisj : set_disj M ((B |_| B') |_| G)).
    apply fasdDisj with h; hyr HpD (sconjAssocL HpD); sconjLift2 HpD; sconjLift3 HpD;
      scm; apply fasdMerge; assumption.
    apply (set_disj_elim MBGBdisj z).
    assumption.
    rewrite <- BBGeqv; assumption.


    (* done with the various lemmas *)
    lli.
    constructor.
    intro H; simpl in H; destruct H as [H | H]; [discriminate H | assumption].
    constructor.
    intuition.
    constructor.
    assumption.
    constructor.
    assumption.
    constructor.
    assumption.
    constructor.
    intros r I; destruct I as [I | I].
    subst r; constructor; auto.
    elf; auto.
    
    (* showing the virtual heap is related to the conc heap by the map *)
    assert (Disj1 : set_disj (B |_| B') (G |_| W)).
    apply fasdDisj with h.
    lle HpD; sconjLift3 HpD; hyr HpD (sconjAssocR HpD); sconjLift2 HpD; scm;
    apply fasdMerge; assumption.
    assert (Disj2 : set_disj G W).
    apply fasdDisj with h; lle HpD; sconjLift3 HpD; scm.
    
    apply fieldMappedEqv with ((B |_| B') |_| G |_| W).
    apply fasMerge.
    sconjLift3of3 HpH; sconjLift3of3 HpH; scm.
    hyr Hp0 (objHpIsObjs Hp0); llMatch; lleim.
    apply fasMerge.
    scm; hyr Hp0 (objHpIsObjs Hp0); llMatch; lleim.
    assumption.
    assumption.
    apply fasMerge.
    2:assumption.
    unfold objMapped; unfold fieldMapped; scm.
    (* case: x in B |_| B' *)
    llMatch; hpunfoldIn Hp0; lle Hp0.
    assert (HQ : h' ||- (okObjAt (B |_| G |_| B') x ** top) ** top).
    sconjLift2 HpH; scm; hyr Hp1 (objHpElim I Hp1); scm.
    hpunfoldIn HQ; lle HQ.
    assert (Sub' : Heap.subHeap h' (heapOf s)).
    apply Heap.subHeapTrans with h.
    apply (sconjEqSubHeap Hp).
    assumption.
    assert (HY := subHeapImpl HQ Sub'); lle HY;
      assert (E1 := contContEq H3 HY); sconjLift2 HY;
	assert (E2 := contContEq H4 HY); subst x3 x2; clear H3 H4 HY.
    lli.
    2:apply Hp0.
    exists x0; rewrite PeqB; [constructor; [reflexivity | scm] | assumption].
    exists x1; rewrite PeqB; [constructor; [reflexivity | 
      sconjLift2 HQ; scm] | assumption].
    (* case: x in G |_| W *)
    hpunfoldIn Hp0; llMatch; lle Hp0.
    destruct H3 as [v0 [HpV0 Meq0]]; destruct H4 as [v1 [HpV1 Meq1]].
    assert (Hp' : h' ||- (x |->__ ** top) ** top).
    destruct (set_union_elim I) as [I' | I']; [sconjLift3of3 HpH | idtac]; scm;
      hyr Hp1 (fasElim I' (objHpIsObjs Hp1)); scm.
    lle Hp'.
    assert (HpX : h ||- Heap.heap_eqv h' ** top).
    scm.
    hyr HpX (sconjInsert HpX Hp'); lle HpX; hyr HpX (subHeapImpl HpX Sub);
    lle HpX; assert (E1 := contContEq HpX HpV0); 
      sconjLift2 HpX; assert (E2 := contContEq HpX HpV1); subst x2 x3; clear HpX Hp.
    scm; lli.
    2:eauto.
    exists v0; constructor.
    unfold mapX in Meq0; assert (IO := isOddOk v0); destruct (isOdd v0).
    injection Meq0; intro; subst x0; rewrite PeqB; [reflexivity | left; assumption].
    unfold map_union in Meq0; destruct (phi v0); injection Meq0; intro; subst; reflexivity.
    scm.
    2:eauto.
    exists v1; constructor.
    unfold mapX in Meq1; assert (IO := isOddOk v1); destruct (isOdd v1).
    injection Meq1; intro; subst x1; rewrite PeqB; [reflexivity | left; assumption].
    unfold map_union in Meq1; destruct (phi v1); injection Meq1; intro; subst; reflexivity.
    sconjLift2 Hp'; scm.
  
    (* finally, show that the heap contains everything properly *)
    scm.
    apply aconjIntro.
    assumption.
    apply heapEqvTrue with h'.
    apply Heap.heap_eqvSymm; apply (eqvPredEqv Hp).
    assumption.
  
  Qed.


  Lemma bakerReprAddField : forall s vs rr A,
    s |||- repr vs (rfileOf s) rr ** A -> forall toSt toEnd x free alloc,
    s |||- 8 + rget s rgcInfo |->, (toSt::toEnd::x::free::alloc::nil) ** top -> 
    forall r, 
      okFieldVal ((rangeObjs toSt free) |_| (rangeObjs alloc toEnd)) (rget s r) ->
      ~In r (zero::I.rgcInfo::rr) ->
      rget vs r = rget s r ->
      s |||- repr vs (rfileOf s) (r::rr) ** A.
  Proof.
    unfold repr, Baker.repr.
    intros s vs rr A Hp toSt0 toEnd0 y free0 alloc0 Hp' r okR NeqR VeqR; lle Hp.
    rename x1 into toSt; rename x3 into scan; rename x4 into free; rename x5 into alloc;
      rename x2 into toEnd; rename x6 into phi; rename x7 into W; rename x8 into M; 
	rename x into frSt; rename x0 into frEnd; rename x9 into vh.
    destruct H3 as [NI [als [alFr [WMeq [Iso [Req VsEq]]]]]].
    lle Hp'.
    assert (toSt0 = toSt) by (sconjLift6 Hp; exact (contContEq Hp' Hp)).
    assert (toEnd0 = toEnd).
      sconjLift7 Hp; sconjLift2 Hp'; exact (contContEq Hp' Hp).
    assert (y = scan).
      sconjLift8 Hp; sconjLift3 Hp'; exact (contContEq Hp' Hp).
    assert (free0 = free).
      sconjLift9 Hp; sconjLift4 Hp'; exact (contContEq Hp' Hp).
    assert (alloc0 = alloc).
      sconjLift10 Hp; sconjLift5 Hp'; exact (contContEq Hp' Hp).
    subst toSt0 toEnd0 y free0 alloc0.
    clear Hp'.
    hyr Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR
      (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp))))))))); scm; lle Hp; clear A.
    existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl scan; 
      existsl free; existsl alloc; existsl phi; existsl W; existsl M; existsl vh.
    lli; [clear Hp | assumption].
    repeat (constructor; [assumption | ]).
    constructor.
    simpl; intros [I | I]; [subst r; apply NeqR; right; left; reflexivity | apply NI; assumption].
    repeat (apply conj; [auto | ]). 2:assumption.
    intros r0 I.
    destruct (regCompare r0 r) as [E | E].
    destruct (regCompare r0 zero) as [E' | E'].
    subst r0 r; elf; apply NeqR; left; reflexivity.
    subst r0; stSimpl.
    constructor; [assumption | apply (okFieldValWeaken okR)].
    assert (AL : aligned8 toSt scan).
      simpl in als; decompose [and] als; assumption.
    assert (LE : scan <= free).
      simpl in als; decompose [and] als; destruct als as [_ [[k X] _]]; omega.
    rewrite <- (set_eqvUnionAssoc (rangeObjs toSt scan) 
      (rangeObjs scan free) (rangeObjs alloc toEnd)); rewrite (rangeObjsTransEqv AL LE).
    apply set_subsetRefl.
    destruct I as [I | I].
    elf; auto.
    stSimpl; apply (Req r0); assumption.
  Qed.


  Lemma bakerStOkAllocPostIsRepr : forall s vs A,
    bakerStOkPost vs A s ->
    let x := rget s v0 in
    let vs' := rset (uhset (uhset vs x 1) (4 + x) 1) v0 x in
      s |||- repr vs' (rfileOf s) (v0 :: rroot :: nil) ** A.
  Proof.
    intros s vs A [frStart [frEnd [toStart [toEnd [scan [free [alloc X]]]]]]].
    destruct X as [rootEq [[v0In [Hneq1 Hneq2]] Bok']].
    intros x vs'.
    assert (HpV' : s |||- repr vs' (rfileOf s) (rroot :: nil) ** A).
    apply bakerStOkIsRepr.
    exists frStart; exists frEnd; exists toStart; exists toEnd;
      exists scan; exists free; exists alloc; stSimpl; constructor; assumption.
    apply bakerReprAddField with toStart toEnd scan free alloc.
    assumption.
    destruct Bok' as [phi [W [M X]]]; intuition; intuition.
    apply sconjComm; apply topTopEL.
    sconjLift6 H4; hyr H4 (sconjComm H4); scm;
    apply topTopEL; simpl in Hp; scm; simpl; scm.
    right; sui2; assumption.
    simpl; intros [I | [I | [I | I]]]; try assumption; discriminate I.
    subst vs'; stSimpl; reflexivity.

  Qed.

  
  Lemma bakerAllocPreOk : forall s, 
    VGC.allocPre repr rroot s -> 
    bakerAllocPre s.
  Proof.
    intros s [vs Hp].
    exists vs; exact (reprIsBakerStOk _ Hp).
  Qed.


  Lemma bakerAllocPreOk' : forall s, 
    bakerAllocPre s -> 
    VGC.allocPre repr rroot s.
  Proof.
    intros s [vs Hp].
    exists vs; exact (bakerStOkIsRepr Hp).
  Qed.


  Lemma bakerAllocPreEqv : forall s,
    bakerAllocPre s <-> VGC.allocPre repr rroot s.
  Proof.
    intro s; constructor; intro H.
    exact (bakerAllocPreOk' H).
    exact (bakerAllocPreOk H).
  Qed.


  Definition bakerAllocGuar' : state -> state -> Prop :=
    VGC.allocGuar repr rroot BMGI.gcGuar BMGI.presReg 1.


  Lemma inclMapEqv : forall f rr (r:rnum),
    incl rr (r::nil) ->
    rr = nil \/ set_eqv (List.map f rr) (set_add (f r) empty_set).
  Proof.
    intros f rr r' I.
    destruct rr as [ |r rr ].
    auto.
    right; destruct (regCompare r r').
    subst r'.
    apply regListMapEqv.
    intros r' II; assert (X := I _ II).
    simpl in X; intuition.
    assert (II : In r (r::rr)).
    left; reflexivity.
    hyr II (I _ II).
    destruct II.
    elf; auto.
    intuition.
  Qed.


  Lemma virtGuarOk : forall vs vs',
    bakerVirtGuar vs vs' ->
    BMGI.gcGuar vs vs' (rroot :: nil).
  Proof.
    unfold bakerVirtGuar, BMGI.gcGuar, BakerGCGuar.gcGuar.
    intros vs vs' Guar objs h Vhp.

    destruct (Guar objs h) as [objs' [phi X]]; clear Guar.
    scm.

    destruct X as [Iso [rootEq Vhp0]].
    destruct (sconjExtract Vhp0) as [h' [Vhp' Hp']]; clear Vhp0.
    exists h'; constructor.
    assumption.

    (* show the isomorphism holds *)
    unfold isoState.
    stSimpl;
    replace (rget (h, rfileOf vs)) with (rget vs) by reflexivity;
      replace (rget (h', rfileOf vs')) with (rget vs') by reflexivity;
    exists objs; exists objs'; exists phi.

    assert (HpOk : h ||- objHp objs objs).
    apply heapEqvImplL with (heapOf vs) top; scm.
    destruct (aconjElim Hp); apply aconjIntro.
    exact (minObjHpElim1 H).
    assumption.
    
    constructor.
    assumption.
    constructor.
    
    eapply objHpIsFwdedH; eauto.
    constructor.
    assumption.
    constructor.
    intros r [E | Ir].  subst r; assumption. elf; auto.

    assumption.
  
  Qed.


  Lemma bakerAllocGuarOk : forall s s',
    bakerAllocGuar s s' ->
    bakerAllocGuar' s s'.
  Proof.
    intros s s'; 
      unfold bakerAllocGuar, bakerAllocGuar', allocGuar;
        intros [Guar [Rok raEq]].
    constructor. 2:assumption.
    
    intros vs A Hp.
    hyr Hp (reprIsBakerStOk _ Hp).
    destruct (Guar vs A Hp) as [vs' [Guar' Hp']]; clear Guar Hp.
    exists vs'; exists (rget s' v0).
    constructor.
    apply virtGuarOk; assumption.
    constructor.
    exact (bakerStOkAllocPostIsRepr Hp').
    
    intros B Hp''; stSimpl.
    replace (heapOf (uhset (uhset vs' (rget s' v0) 1) (4 + rget s' v0) 1)) with
      (Heap.uset (Heap.uset (heapOf vs') (rget s' v0) 1) (4 + rget s' v0) 1)
      by reflexivity.
    apply sconjComm; lli; apply sconjGLift2'.
    unfold bakerStOkPost in Hp'.
    destruct Hp' as [frS [frE [toS [toE [scan [free [alloc [rootEq [v0In Bok]]]]]]]]].
    destruct v0In as [v0In [GE1 GE2]].
    destruct Bok as [phi [W [M [Al1 [Al2 [Wmeq [Iso [rootOk [HpOk Vok]]]]]]]]].
    generalize dependent Vok; stSimpl; intro.

    assert (PP : ptrP (rget s' v0)).
      sconjLift3 HpOk.
      assert (HX : s' |||- (rget s' v0 |->__ ** top) ** top).
        scm.
        assert (I : set_In (rget s' v0) (rangeObjs toS scan |_| rangeObjs alloc toE)).
          sui2; assumption.
        hyr Hp (objHpElim I Hp); hpunfoldIn Hp; scm; lleim.
      lle HX; apply (contElim1L HX).

    apply heapAddOk. apply heapAddOk. assumption. assumption.
    intro I; lle I; assert (X := contGetEqL I).
    rewrite X in GE1; discriminate GE1.
    constructor; auto.
    intro I; lle I; assert (X := contGetEqL I).
    replace (Heap.get (Heap.uset (heapOf vs') (rget s' v0) 1) (4 + rget s' v0)) with
      (hget (uhset vs' (rget s' v0) 1) (4 + rget s' v0)) in X by reflexivity.
    rewrite hgetNeq in X.
    replace (hget vs') with (Heap.get (heapOf vs')) in X by reflexivity.
    rewrite X in GE2; discriminate GE2.
    constructor; auto.
    assumption.
    omega.
  Qed.



  Lemma bakerAllocOk : iseqOkS Pbaker (VGC.allocPre repr rroot) bakerAllocGuar' bakerAlloc.
  Proof.
    eapply iseqOkWeak.
    apply bakerAllocOk.
    intros s Pre; constructor.
    exact (bakerAllocPreOk Pre).
    exact (bakerAllocGuarOk (s := s)).
  Qed.

  Notation bakerAllocSpec := (bspeci (VGC.allocPre repr rroot) bakerAllocGuar').


  Definition phi n := 
    match Pbaker n with
      | SomeT x => SomeT x
      | NoneT => match Pscanf n with
		   | SomeT x => SomeT x
		   | NoneT =>
		     if beq_nat n I.allocLbl then
		       SomeT bakerAllocSpec
		     else if beq_nat n I.read1Lbl then
		       SomeT (bakerReadSpec 0)
		     else if beq_nat n I.read2Lbl then
		       SomeT (bakerReadSpec 4)
		     else if beq_nat n BAKERREADRET then
		       SomeT bakerReadRetSpec
		     else if beq_nat n I.write1Lbl then
		       SomeT (bakerWriteSpec 0)
		     else if beq_nat n I.write2Lbl then
		       SomeT (bakerWriteSpec 4)
		     else
		       NoneT _
		 end
    end.


  Ltac phiStep Ne f k :=
    destruct (natEqDec f k) as [R | R];
      [subst f; rewrite beq_true in Ne; apply setRangeIntro; omega |
       rewrite (beq_false R) in Ne; clear R].

  Lemma gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.
  Proof.
    unfold I.gcPhiDom; unfold phi; unfold Pbaker; unfold Pscanf.
    unfold SCANFIELD; unfold BAKERLOOP; unfold BAKERLOOP2; unfold BAKEREXIT;
      unfold RESTOREROOT; unfold INFLOOP.
    unfold RETURN; unfold SCANPOINTER; unfold NOTTOSPACEPTR; unfold NOTFROMSPACEPTR; 
      unfold SCANCOPIED; unfold SCANNOCOPY; unfold FWDOBJ; 
	unfold SCANFIELD2; unfold RRETURN.
    unfold I.allocLbl; unfold I.read1Lbl; unfold I.read2Lbl; unfold BAKERREADRET;
      unfold I.write1Lbl; unfold I.write2Lbl.
    intros f Ne.
    phiStep Ne f 89.
    phiStep Ne f 31.
    phiStep Ne f 32.
    phiStep Ne f 33.
    phiStep Ne f 34.
    phiStep Ne f 29.
    phiStep Ne f 88.
    phiStep Ne f 26.
    phiStep Ne f 24.
    phiStep Ne f 25.
    phiStep Ne f 27.
    phiStep Ne f 28.
    phiStep Ne f 23.
    phiStep Ne f 21.
    phiStep Ne f 22.
    phiStep Ne f 10.
    phiStep Ne f 20.
    phiStep Ne f 30.
    phiStep Ne f 38.
    phiStep Ne f 40.
    phiStep Ne f 50.
    elf; auto.
  Qed.


  Ltac PhiOkSolver :=
    unfold phi; unfold Pbaker; unfold Pscanf; rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    simpl; intuition.


  Lemma read1PhiOk :
    optPredT (phi I.read1Lbl)
      (specWeaker (read1Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma read2PhiOk :
    optPredT (phi I.read2Lbl)
      (specWeaker (read2Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write1PhiOk : 
    optPredT (phi I.write1Lbl)
      (specWeaker (write1Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write2PhiOk : 
    optPredT (phi I.write2Lbl)
      (specWeaker (write2Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma allocPhiOk : 
    optPredT (phi I.allocLbl)
      (specWeaker (allocSpec repr I.rroot I.gcGuar I.presReg I.NULL)).
  Proof.
    PhiOkSolver.
  Qed.


  Definition Gbarr f :=
    if beq_nat f I.allocLbl then
      Some bakerAlloc
    else if beq_nat f I.read1Lbl then
      Some (bakerRead 0)
    else if beq_nat f I.read2Lbl then
      Some (bakerRead 4)
    else if beq_nat f BAKERREADRET then
      Some bakerReadRet
    else if beq_nat f I.write1Lbl then
      Some (bakerWrite 0)
    else if beq_nat f I.write2Lbl then
      Some (bakerWrite 4)
    else
      None.

  Definition Gscan f :=
    if beq_nat f RETURN then
      Some returnBlock
    else if beq_nat f SCANPOINTER then
      Some scanPointer
    else if beq_nat f NOTTOSPACEPTR then
      Some isNotToSpacePtr
    else if beq_nat f NOTFROMSPACEPTR then
      Some isNotFromSpacePtr
    else if beq_nat f SCANCOPIED then
      Some scanCopied
    else if beq_nat f SCANNOCOPY then
      Some scanNoCopy
    else if beq_nat f FWDOBJ then
      Some fwdObj
    else if beq_nat f SCANFIELD2 then
      Some scanField2
    else if beq_nat f RRETURN then
      Some rReturn
    else
      None.

  Definition Ggc f :=
    if beq_nat f SCANFIELD then
      Some scanField
    else if beq_nat f BAKERLOOP then
      Some bakerLoop
    else if beq_nat f BAKERLOOP2 then
      Some bakerLoop2
    else if beq_nat f BAKEREXIT then
      Some bakerExit
    else if beq_nat f RESTOREROOT then
      Some restoreRoot
    else if beq_nat f INFLOOP then
      Some infLoop
    else
      None.

  Definition GCC := map_union Ggc (map_union Gscan Gbarr).
  

  Ltac weakerStep f f' H p g :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true;
	injection 1; clear H; intros; subst p g;
	  econstructor; econstructor; constructor;
	    [unfold Pbaker; unfold Pscanf; repeat (rewrite beq_false; [idtac | discriminate]) | 
	      constructor; [reflexivity | intros s s' Guar; apply Guar]] |
       rewrite (beq_false R); clear R].


  Lemma phiWeakerPscanf : forall f p0 g0,
    Pscanf f = SomeT (bspeci p0 g0) ->
    exists p1, exists g1, phi f = SomeT (bspeci p1 g1) /\
      p0 = p1 /\ forall s s', g1 s s' -> g0 s s'.
  Proof.
    unfold ScanFieldB.Pscanf; unfold phi.

    intros f p g.

    weakerStep f RETURN H p g.
    reflexivity.
    weakerStep f SCANPOINTER H p g.
    reflexivity.
    weakerStep f NOTTOSPACEPTR H p g.
    reflexivity.
    weakerStep f NOTFROMSPACEPTR H p g.
    reflexivity.
    weakerStep f SCANCOPIED H p g.
    reflexivity.
    weakerStep f SCANNOCOPY H p g.
    reflexivity.
    weakerStep f FWDOBJ H p g.
    reflexivity.
    weakerStep f SCANFIELD2 H p g.
    reflexivity.
    weakerStep f SCANFIELD H p g.
    reflexivity.

    destruct (natEqDec f RRETURN) as [R | R].
    subst f; rewrite beq_true.
    injection 1; clear H; intros; subst p g.
    econstructor; econstructor; constructor.
    unfold Pbaker; unfold Pscanf; repeat (rewrite beq_false; [idtac | discriminate]).
    rewrite beq_true.
    reflexivity.
    constructor; [reflexivity | intros s s' Guar; apply Guar].
    rewrite (beq_false R); clear R.

    destruct (natEqDec f INFLOOP) as [R | R].
    subst f; rewrite beq_true.
    injection 1; clear H; intros; subst p g.
    econstructor; econstructor; constructor.
    unfold Pbaker; unfold Pscanf; repeat (rewrite beq_false; [idtac | discriminate]).
    rewrite beq_true.
    reflexivity.
    constructor; [reflexivity | intros s s' Guar; apply Guar].
    rewrite (beq_false R); clear R.

    discriminate 1.

  Qed.


  Lemma phiWeakerPbaker : forall f p g,
    Pbaker f = SomeT (bspeci p g) ->
    exists p', exists g', phi f = SomeT (bspeci p' g') /\
      p = p' /\ forall s s', g' s s' -> g s s'.
  Proof.
    unfold phi.
    intros f p g Peq; destruct (Pbaker f).
    injection Peq; intro; subst b; clear Peq.
    econstructor; econstructor; constructor.
    reflexivity.
    intuition.
    discriminate.
  Qed.


  Ltac GCCokStep' Peq f f' p g :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true; injection 1; clear Peq; intros; subst p g |
	rewrite (beq_false R); let R0 := fresh "R" in rename R into R0].

  Ltac GCCokStep Peq f f' p g :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true; injection 1; clear Peq; intros; subst p g |
	rewrite (beq_false R); clear R].
    
  Lemma GCCok : cdHeapOkSC phi GCC phi.
  Proof.
    intros f [p g] Peq.
    unfold GCC, map_union, Ggc in * .
      (* previous line kills Proof General with the inf loop problem. *)
    unfold phi, Pbaker in Peq; generalize dependent Peq.

    GCCokStep' Peq f SCANFIELD p g.
    2:GCCokStep Peq f BAKERLOOP p g.
    3:GCCokStep Peq f BAKERLOOP2 p g.
    4:GCCokStep Peq f BAKEREXIT p g.
    5:GCCokStep Peq f RESTOREROOT p g.
    6:GCCokStep' Peq f INFLOOP p g.

    apply (iseqOkWeakP _ _ _ _ scanFieldOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ bakerLoopOk _ phiWeakerPbaker).
    apply (iseqOkWeakP _ _ _ _ bakerLoop2Ok _ phiWeakerPbaker).
    apply (iseqOkWeakP _ _ _ _ bakerExitOk _ phiWeakerPbaker).
    apply (iseqOkWeakP _ _ _ _ restoreRootOk _ phiWeakerPbaker).
    apply (iseqOkWeakP _ _ _ _ infLoopOk _ phiWeakerPscanf).

    unfold Pscanf; unfold Gscan.
    GCCokStep Peq f RETURN p g.
    2:GCCokStep Peq f SCANPOINTER p g.
    3:GCCokStep Peq f NOTTOSPACEPTR p g.
    4:GCCokStep Peq f NOTFROMSPACEPTR p g.
    5:GCCokStep Peq f SCANCOPIED p g.
    6:GCCokStep Peq f SCANNOCOPY p g.
    7:GCCokStep Peq f FWDOBJ p g.
    8:GCCokStep Peq f SCANFIELD2 p g.
    9:rewrite (beq_false R0); GCCokStep Peq f RRETURN p g.

    eapply iseqOkWeakP; [apply CheneyLemma.returnOk | discriminate 1].
    apply (iseqOkWeakP _ _ _ _ scanPointerOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ isNotToSpacePtrOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ isNotFromSpacePtrOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanCopiedOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanNoCopyOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ fwdObjOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanField2Ok _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ rReturnOk _ phiWeakerPscanf).

    rewrite (beq_false R1); clear R0 R1; unfold Gbarr.
    GCCokStep' Peq f I.allocLbl p g.
    2:GCCokStep Peq f I.read1Lbl p g.
    3:GCCokStep Peq f I.read2Lbl p g.
    4:GCCokStep Peq f BAKERREADRET p g.
    5:GCCokStep Peq f I.write1Lbl p g.
    6:GCCokStep Peq f I.write2Lbl p g.

    apply (iseqOkWeakP _ _ _ _ bakerAllocOk _ phiWeakerPbaker).
    apply bakerReadOk; auto.
    apply bakerReadOk; auto.
    apply bakerReadRetOk.
    apply bakerWriteOk; auto.
    apply bakerWriteOk; auto.
    discriminate .

  Qed.


End BakerGCImpl.
