Require Export GC.
Require Export LinLog.
Require Export State.
Require Export List.
Require Export VGC.

Require Import StateTheory.


Set Implicit Arguments.


(* things pertaining to copying GC *)

Definition isoState rr s s' :=
  exists objs, exists objs', exists phi,
    s |||- objHp objs objs /\
    s' |||- objHp objs' objs' /\
    iso_set phi objs objs' /\
    (forall r, In r rr -> mapX phi (rget s r) = Some (rget s' r)) /\
    s' |||- (foralls x in objs', objIsFwdedH phi (heapOf s) x).


Module CopyGCGuar <: BASIC_MGI_ARG.
  Definition gcGuar (s s':state) (rr:list rnum) := 
    forall objs h,
      s |||- (minObjHp (List.map (rget s) rr) objs && Heap.heap_eqv h) ** top ->
	exists h',
	  s' |||- Heap.heap_eqv h' ** top /\
	  isoState rr (h, rfileOf s) (h', rfileOf s').

End CopyGCGuar.

Module CopyMGI := BasicMGI CopyGCGuar.


Lemma rootReachableSubheap : forall h objs,
  h ||- objHp objs objs -> forall h',
  Heap.subHeap h h' -> forall v,
  okFieldVal objs v -> forall x,
  rootReachable h' (v::nil) x ->
  rootReachable h (v::nil) x.
Proof.
  intros h objs Hp h' S v Vok x [b [I R]].
  simpl in I; destruct I.
  2:elimtype False; assumption.
  subst b.
  exists v; constructor.
  left; reflexivity.

  generalize dependent v.
  induction 2.
  constructor.

  destruct Vok as [Odd | I].

  assert (Q := isPtrOk x); unfold Heap.get in H; destruct (isPtr x).
  elimtype False; eapply not_even_and_odd.
  apply ptrPeven; eauto.
  assumption.
  discriminate H.

  hypReplace Hp (objHpElim I Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  rewrite (S x x0 (contGetEqL Hp)) in H; injection H; clear H; intro; subst x0.
  econstructor 2.
  instantiate (1 := x').
  eapply contGetEqL; eauto.
  apply IHR.
  assumption.

  destruct Vok as [Odd | I].

  assert (Q := isPtrOk (4 + x)); unfold Heap.get in H; destruct (isPtr (4 + x)).
  elimtype False; eapply not_even_and_odd.
  apply ptrPeven; eauto.
  simpl; constructor; constructor; constructor; constructor; assumption.
  discriminate H.

  hypReplace Hp (objHpElim I Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  sconjLift2 Hp.
  rewrite (S _ _ (contGetEqL Hp)) in H; injection H; clear H; intro; subst x1.
  econstructor 3.
  instantiate (1 := x').
  eapply contGetEqL; eauto.
  apply IHR.
  assumption.

Qed.



Lemma isoStateAtomElim' : forall rr s s',
  isoState rr s s' -> forall r,
    In r rr ->
    (atomP (rget s r) \/ atomP (rget s' r)) ->
    rget s r = rget s' r.
Proof.
  unfold isoState.
  intros rr s s' [objs [objs' [phi [_ [Hp [Iso [Reqv _]]]]]]] r Irr.
  hyr Reqv (Reqv _ Irr); unfold mapX in Reqv; clear Irr.
  intros [Atom | Atom].

  rewrite (isOddTrue Atom) in Reqv.
  injection Reqv; auto.

  assert (IO := isOddOk (rget s r)).
  destruct (isOdd (rget s r)).
  injection Reqv; auto.
  elf; apply (not_even_and_odd (rget s' r)).
  2:assumption.
  apply ptrPeven.
  assert (Ir : set_In (rget s' r) objs').
  destruct Iso as [_ [Iso _]].
  eauto.
  hyr Hp (objHpElim Ir Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp; apply (contElim1L Hp).
Qed.



Lemma isoStateAtomElim : forall rr s s',
  isoState rr s s' -> forall r,
    In r rr ->
    (atomP (rget s r) <-> atomP (rget s' r)).
Proof.
  intros rr s s' Iso r Irr; constructor; intro Atom.

  rewrite <- (isoStateAtomElim' Iso _ Irr).
  assumption.
  auto.

  rewrite (isoStateAtomElim' Iso _ Irr).
  assumption.
  auto.

Qed.


Lemma isoStateCopyReg : forall rr s s',
  isoState rr s s' -> forall rs,
    In rs rr -> forall rd,
      rd <> zero ->
      isoState (rd::rr) (rset s rd (rget s rs)) (rset s' rd (rget s' rs)).
Proof.
  unfold isoState.
  intros rr s s' [objs [objs' [phi [Hp [Hp' [Iso [Req Fwd]]]]]]] rs Irr rd NZ.
  exists objs; exists objs'; exists phi; stSimpl.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  2:assumption.

  simpl; intros r [E | Irr'].
  subst r; stSimpl; auto.
  destruct (regCompare r rd).
  subst r; stSimpl; auto.
  stSimpl.
  auto.
Qed.


Lemma isoStatePtr : forall rr s s',
  isoState rr s s' -> forall r,
    In r rr -> forall k, k = 0 \/ k = 4 ->
    s |||- k + rget s r |->_ ** top ->
      s' |||- k + rget s' r |->_ ** top.
Proof.
  unfold isoState.
  intros rr s s' [objs [objs' [phi [Hp [Hp' [Iso [Req _]]]]]]]
    r Irr k Keq HpR.
  hyr Req (Req _ Irr); clear Irr.
  unfold mapX in Req; rewrite isOddFalse in Req.
  assert (Ir : set_In (rget s' r) objs').
  destruct Iso as [_ [Iso _]].
  eauto.
  hyr Hp' (objHpElim Ir Hp').
  unfold okObjAt in Hp'; unfold okFieldAt in Hp'; lle Hp'; destruct Keq; subst k;
    [idtac | sconjLift2 Hp']; lli; scm.
  clear Req.
  lle HpR; assert (X := contElim1L HpR).
  apply ptrPeven; destruct Keq; subst k.
  assumption.
  inversion X; auto.
Qed.


Lemma isoStateRootWeaken : forall rr s s',
  isoState rr s s' -> forall rr',
    incl rr' rr ->
    isoState rr' s s'.
Proof.
  intros rr s s' [objs [objs' [phi [Hp [Hp' [Iso [Req Fwd]]]]]]].
  exists objs; exists objs'; exists phi.
  repeat (constructor; [assumption | idtac]).
  constructor; [idtac | assumption].
  intros r Irr; apply Req; auto.
Qed.


Lemma isoStateAddAtomRoot : forall rr vs vs',
  isoState rr vs vs' -> forall n,
  atomP n -> forall r,
    ~In r (zero::rr) ->
    isoState (r::rr) (rset vs r n) (rset vs' r n).
Proof.
  intros rr vs vs' [objs [objs' [phi [Hp [Hp' [Iso [Req Fwd]]]]]]] n Atom r Ni.
  exists objs; exists objs'; exists phi.
  repeat (constructor; [assumption | idtac]).
  constructor; [idtac | assumption].
  destruct (regCompare r zero).
  elf; subst r; apply Ni; left; auto.
  simpl; intros r' [E | Irr].
  subst r'.
  stSimpl; unfold mapX; rewrite (isOddTrue Atom); reflexivity.
  destruct (regCompare r r').
  elf; subst r'; apply Ni; right; auto.
  stSimpl.
  auto.
Qed.


Lemma isoPresPtr : forall rr vs vs',
  isoState rr vs vs' -> forall r,
  In r rr ->
  vs' |||- rget vs' r |->_ ** top ->
  vs |||- rget vs r |->_ ** top.
Proof.
  intros rr vs vs' [objs [objs' [phi [Hp [Hp' [Iso [Req Fwd]]]]]]] r Irr Hp''.
  hyr Req (Req _ Irr).
  assert (IO := isOddOk (rget vs r)); unfold mapX in Req;
    destruct (isOdd (rget vs r)).

  elf; injection Req; clear Req; intro E; apply (not_even_and_odd (rget vs r)).
  apply ptrPeven; rewrite E; lle Hp''; apply (contElim1L Hp'').
  assumption.

  assert (Ir : set_In (rget vs r) objs).
  destruct Iso as [Iso _]; destruct (Iso (rget vs r)) as [_ Q]; apply Q;
    rewrite Req; discriminate.
  hyr Hp (objHpElim Ir Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp; lli; scm.
Qed.


(* ???
Lemma copyGuarPresPtr : forall rr vs vs',
  CopyMGI.gcGuar vs vs' rr -> forall vs'',
  isoState rr vs'' vs -> forall r,
    In r rr ->
    vs' |||- rget vs' r |->_ ** top ->
    vs |||- rget vs r |->_ ** top.
Proof.
  unfold CopyMGI.gcGuar; unfold CopyGCGuar.gcGuar.
  intros rr vs vs' Guar vs'' Iso1 r Irr Hp''.

  assert (HpX : vs |||- (lexists objs', minObjHp (List.map (rget vs) rr) objs') ** top).
  destruct Iso1 as [objs [objs' [phi [Hp [Hp' [Iso [Req Fwd]]]]]]].
  apply objHpHasMin with objs'.
  assumption.
  intros r' Irr'.
  hyr Req (Req _ Irr').
  assert (IO := isOddOk (rget vs'' r')); unfold mapX in Req;
    destruct (isOdd (rget vs'' r')).
  injection Req; clear Req; intro E; rewrite <- E; left; assumption.
  right; destruct Iso as [_ [Iso _]]; eauto.
  lle HpX; rename x into objs''.
  destruct (sconjExtract HpX) as [mh [Hp2 Vhp]]; clear HpX.
  assert (Hp2' : vs |||- (minObjHp (List.map (rget vs) rr) objs'' && Heap.heap_eqv mh) ** top).
  scm.
  apply aconjIntro.
  apply heapEqvTrue with mh.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; auto.
  auto.
  auto.

  assert (X : incl rr rr).
  intro rx; auto.
  hyr Guar (Guar _ _ _ X Hp2'); clear X Hp2'.
  destruct Guar as [h' [VHp [objs [objs' [phi [Hp [Hp' [Iso [Req Fwd]]]]]]]]].
  simpl in Fwd; simpl in Hp; simpl in Hp'.
  hyr Req (Req _ Irr).
  replace (rget (mh, rfileOf vs) r) with (rget vs r) in Req; try reflexivity.
  replace (rget (h', rfileOf vs') r) with (rget vs' r) in Req; try reflexivity.
  assert (IO := isOddOk (rget vs r)); unfold mapX in Req; destruct (isOdd (rget vs r)).
  elf; injection Req; intro; apply (not_even_and_odd (rget vs r)); try assumption.
  apply ptrPeven; rewrite H; lle Hp''; apply (contElim1L Hp'').
  assert (II : set_In (rget vs r) objs).
  destruct Iso as [Iso _]; destruct (Iso (rget vs r)) as [_ Q];
    apply Q; rewrite Req; discriminate.
  hyr Hp (objHpElim II Hp); unfold okObjAt in Hp; unfold okFieldAt in Hp;
    lle Hp; hyr Hp2 (sconjInsert Hp2 Hp); lle Hp2; lli; scm.
Qed.
*)


(*
Lemma listMapSubset : forall (X Y:list rnum),
  incl X Y -> forall f,
    set_subset (List.map f X) (List.map f Y).
Proof.
  induction X; intros Y II f x Ix.
  elf; apply Ix.
  simpl in Ix; destruct Ix.

  subst x.
  assert (Iy : In a Y).
  apply (II a); left; reflexivity.
  clear II IHX X; induction Y.
  apply Iy.
  simpl in Iy; destruct Iy as [Iy | Iy].
  subst a0; simpl; auto.
  right; auto.
  
  apply IHX; clear IHX.
  intros z Iz; apply II; right; assumption.
  assumption.

Qed.

Lemma isoStepCommute : forall rr vs s,
  isoState rr vs s -> forall s',
    CopyMGI.gcGuar s s' rr ->
    exists vs', BasicGCMGI.gcGuar vs vs' rr /\ isoState rr vs' s'.
Proof.
  intros rr vs s [vobjs [objs [phi [Vhp [Hp [Iso [Req HpFwd]]]]]]];
    unfold CopyMGI.gcGuar; unfold CheneyGCGuar.gcGuar;
      intros s' G.
  
  Axiom objHpHasMin : forall s objs,
    s |||- objHp objs objs -> forall rr,
      (forall r, In r rr -> okFieldVal objs (rget s r)) ->
      s |||- (lexists objs', minObjHp (List.map (rget s) rr) objs') ** top.

  assert (vrootsOk : forall r, In r rr -> okFieldVal vobjs (rget vs r)).
  intros r Irr.
  hyr Req (Req r Irr); unfold mapX in Req; assert (IO := isOddOk (rget vs r));
    destruct (isOdd (rget vs r)); [left | right].
  assumption.
  destruct Iso as [Iso _]; destruct (Iso (rget vs r)) as [_ X]; apply X;
    rewrite Req; discriminate.

  assert (HpMin0 := objHpHasMin _ Vhp _ vrootsOk).
  lle HpMin0; rename x into mvobjs.
  destruct (sconjExtract HpMin0) as [mvh [VhpMin VhpMinOk]]; clear HpMin0.
  exists (mvh, rfileOf vs).
  constructor.

  (* show the guar holds *)
  constructor.
  2:reflexivity.

  intros mobjs mh rr' Irr' Vhp'; stSimpl.

  Axiom minObjHpMinSub : forall h vv objs,
  h ||- minObjHp vv objs -> forall h' vv' objs',
  h' ||- minObjHp vv' objs' ->
  set_subset vv' vv ->
  h ||- Heap.heap_eqv h' ** top.

  eapply minObjHpMinSub; eauto.
  apply (heapEqvImplL Vhp').
  apply listMapSubset; assumption.

  (* show the isoState *)
  assert (rootsOk : forall r, In r rr -> okFieldVal objs (rget s r)).
  intros r Irr.
  hyr Req (Req r Irr); unfold mapX in Req; assert (IO := isOddOk (rget vs r));
    destruct (isOdd (rget vs r)); [left | right].
  injection Req; intro E; rewrite <- E; assumption.
  destruct Iso as [_ [Iso _]]; eauto.

  assert (HpMin0 := objHpHasMin _ Hp _ rootsOk).
  lle HpMin0; rename x into mobjs.
  destruct (sconjExtract HpMin0) as [mh [HpMin HpMinOk]]; clear HpMin0.
  destruct (G mobjs mh rr) as [mh' [Hp' Iso']]; clear G.
  intros z Iz; assumption.
  scm; apply aconjIntro.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; apply Hp0.
  assumption.
  assumption.

  (* used the guar *)
  unfold isoState; stSimpl.

why isn't the range of the isoState the reachable part of s'?

*)


(*
Axiom isoStatePtrReg' : forall rr s s',
  isoState rr s s' -> forall r,
    In r rr -> forall r',
      In r' rr ->
      s' |||- rget s' r |-> rget s' r' ** top ->
	s |||- rget s r |-> rget s r' ** top.
*)
