Require Omega.
Require Export EqNat.
Require Export Coq.Arith.Div2.
Require Export Even.
Require Import Bool.


Ltac elf := elimtype False.


Set Implicit Arguments.


Definition negb := negb.

(* instead use negb
Definition bnot (b:bool) : bool :=
  if b then
    false
  else
    true.
*)


Fixpoint isOdd (m:nat) : bool :=
  match m with
    | 0 => false
    | S n => isEven n
  end
with isEven (m:nat) : bool :=
  match m with
    | 0 => true
    | S n => isOdd n
  end.

Fixpoint blt_nat (m n:nat) {struct n} : bool :=
  match n with
    | 0 => false
    | S n' =>
      match m with
	| 0 => true
	| S m' => blt_nat m' n'
      end
  end.

Definition min_nat m n :=
  if blt_nat m n then m else n.


(* the equality for beq_nat_refl is in the wrong order for rewrite tactic *)
Lemma beq_true : forall m,
  beq_nat m m = true.
Proof.
  intro.
  assert (X := beq_nat_refl m).
  symmetry.
  assumption.
Qed.


Lemma beq_false : forall m n,
  m <> n -> beq_nat m n = false.
Proof.
  induction m.
    induction n.
      intuition.
      intuition.
    induction n.
      intuition.
      simpl; intuition.
Qed.


Lemma beq_natSound : forall m n,
  if beq_nat m n then
    m = n
  else
    m <> n.
Proof.
  intros m n; compare m n; intro NE.
  subst m; rewrite beq_true; reflexivity.
  rewrite (beq_false NE); assumption.
Qed.


Lemma blt_true : forall m n,
  m < n -> blt_nat m n = true.
Proof.
  induction m.
    induction n.
      inversion 1.
      auto.
    induction n.
      inversion 1.
      simpl; intros; eapply IHm; omega.
Qed.


Lemma blt_false : forall m n,
  m >= n -> blt_nat m n = false.
Proof.
  induction m.
  induction n.
  auto.
  inversion 1.
  induction n.
  auto.
  simpl; intro; apply IHm; intuition.
Qed.


Lemma ltIsBlt : forall m n,
  blt_nat m n = true -> m < n.
Proof.
  intros m n.
  assert ((m < n)\/(m >= n)).
  omega.
  destruct H.
  auto.
  rewrite (blt_false H).
  discriminate 1.
Qed.


Lemma blt_natSound : forall x y,
  if blt_nat x y then
    x < y
  else
    x >= y.
Proof.
  intros x y.
  assert (D : x < y \/ x >= y).
  omega.
  destruct D as [D | D]; [rewrite (blt_true D) | rewrite (blt_false D)]; assumption.
Qed.


(* Now we define natEqDec non-opaquely *)

Lemma zrefl : 0 = 0.
Proof.
  auto.
Qed.

Lemma sneq : forall n, 0 <> S n.
Proof.
  auto.
Qed.

Lemma sneq2 : forall n, S n <> 0.
Proof.
  auto.
Qed.

Lemma succImpl : forall x y, x = y -> S x = S y.
Proof.
  auto.
Qed.

Lemma succNImpl : forall x y, x <> y -> S x <> S y.
Proof.
  auto.
Qed.

Fixpoint natEqDec (x:nat) : forall y, {x = y} + {x <> y} :=
  match x return forall y, {x = y} + {x <> y} with
    | O => fun y =>
      match y return {O = y} + {O <> y} with
	| O => left _ zrefl
	| S y' => right _ (sneq (n:=y'))
      end
    | S x' => fun y =>
      match y return {S x' = y} + {S x' <> y} with
	| O => right _ (sneq2 (n:=x'))
	| S y' => 
	  match natEqDec x' y' with
            | left P => left _ (succImpl P)
	    | right P => right _ (succNImpl P)
	  end
      end
  end.


Lemma natEqDecRefl : forall x,
  ex (fun P => natEqDec x x = left _ P).
Proof.
  induction x; simpl.
  
  econstructor; reflexivity.

  destruct IHx as [P IH].
  econstructor.
  rewrite IH.
  reflexivity.

Qed.


(* integer division *)
Definition div4 (m:nat) : nat := div2 (div2 m).
Definition div8 x := div2 (div2 (div2 x)).


(* based on ind_0_1_SS from Coq.Arith.Div2 *)
Lemma ind_4 :
 forall P:nat -> Prop,
   P 0 -> P 1 -> P 2 -> P 3 ->
   (forall n, P n -> P (4 + n)) -> forall n, P n.
Proof.
  intros.
  cut (forall n, P n /\ P (1 + n) /\ P (2 + n) /\ P (3 + n)).
  intro Hi; assert (HiN := Hi n); intuition.
  clear n.
  induction n.
  simpl; intuition.
  intuition.
  replace (3 + S n) with (4 + n).
  auto.
  reflexivity.
Qed.


(* based on ind_0_1_SS from Coq.Arith.Div2 *)
Lemma ind_8 :
 forall P:nat -> Prop,
   P 0 -> P 1 -> P 2 -> P 3 -> P 4 -> P 5 -> P 6 -> P 7 ->
   (forall n, P n -> P (8 + n)) -> forall n, P n.
Proof.
  intros.
  cut (forall n, P n /\ P (1 + n) /\ P (2 + n) /\ P (3 + n) /\ P (4 + n) /\ 
    P (5 + n) /\ P (6 + n) /\ P (7 + n)).
  intro Hi; assert (HiN := Hi n); intuition.
  clear n.
  induction n.
  simpl; intuition.
  intuition.
  replace (7 + S n) with (8 + n).
  apply H7; assumption.
  reflexivity.
Qed.


(* pointers are divisible by 4 *)

Fixpoint isPtr (m:nat) : bool :=
  match m with
    | 0 => true
    | S (S (S (S m'))) => isPtr m'
    | _ => false
  end.

Inductive ptrP : nat -> Prop :=
  | nullPtr : ptrP 0
  | succPtr : forall p,
    ptrP p ->
    ptrP (4 + p).


Lemma isPtrOk : forall p,
  if isPtr p then ptrP p else ~ptrP p.
Proof.
  refine (ind_4 (fun p => if isPtr p then ptrP p else ~ptrP p) _ _ _ _ _); simpl.
  constructor.
  intro H; inversion H.
  intro H; inversion H.
  intro H; inversion H.
  intros n IH.
  replace (S (S (S (S n)))) with (4 + n).
  destruct (isPtr n).
  constructor; assumption.
  intro H; apply IH; inversion H; assumption.
  reflexivity.
Qed.


Lemma isPtr_true : forall p,
  ptrP p -> 
  isPtr p = true.
Proof.
  induction 1; simpl; auto.
Qed.


Definition isPtrEq := isPtr_true. (* old name *)

Lemma isPtr_false : forall p,
  ~ptrP p ->
  isPtr p = false.
Proof.
  intros p NP.
  assert (X := isPtrOk p).
  destruct (isPtr p).
  elf; auto.
  reflexivity.
Qed.


Lemma ptrP_elim : forall n,
  ptrP n ->
  exists k, n = 4 * k.
Proof.
  induction 1.
  exists 0; reflexivity.
  destruct IHptrP as [k E].
  subst.
  exists (S k); omega.
Qed.


Lemma ptrP_intro : forall k n,
  n = 4 * k ->
  ptrP n.
Proof.
  induction k; intros n R; rewrite R; clear R.
  constructor.
  replace (4 * S k) with (4 + 4 * k).
  constructor.
  apply IHk.
  reflexivity.
  omega.
Qed.


Lemma ptrP_intro' : forall n,
  ptrP (4 * n).
Proof.
  intro n; apply ptrP_intro with n; reflexivity.
Qed.


Lemma ptrPeven : forall x,
  ptrP x ->
  even x.
Proof.
  induction 1.
  constructor.
  simpl.
  repeat constructor.
  assumption.
Qed.


Lemma ptrPlusPtr : forall x y z,
  4 * x + y = 4 * z ->
  exists k, y = 4 * k.
Proof.
  induction x; intros y z E.
  exists z; omega.

  destruct (IHx (4 + y) z) as [k E'].
  omega.
  destruct k.
  elimtype False; omega.
  exists k; omega.

Qed.
Implicit Arguments ptrPlusPtr [x y z].


Lemma div4plus : forall n, 
  div4 (4 + n) = 1 + div4 n.
Proof.
  apply (ind_4 (fun n => div4 (4 + n) = 1 + div4 n)); 
    unfold div4; simpl; try omega.
  reflexivity.
Qed.


Lemma div8plus : forall n, 
  div8 (8 + n) = 1 + div8 n.
Proof.
  apply (ind_8 (fun n => div8 (8 + n) = 1 + div8 n)); 
    unfold div8; simpl; try omega.
  reflexivity.
Qed.


Lemma div8sub : forall n,
  div8 (n - 8) = div8 n - 1.
Proof.
  apply (ind_8 (fun n => div8 (n - 8) = div8 n - 1)); intros; try reflexivity.
  replace (8 + n - 8) with n.
  replace (div8 (8 + n) - 1) with (div8 n).
  reflexivity.
  intuition.
  intuition.
Qed.


Lemma div44eq : forall x,
  div4 (4 * x) = x.
Proof.
  induction x.
  reflexivity.
  replace (4 * S x) with (4 + 4 * x).
  rewrite div4plus.
  rewrite IHx.
  reflexivity.
  omega.
Qed.


Lemma notIsEvenIsOdd : forall n,
  negb (isEven n) = isOdd n /\ negb (isOdd n) = isEven n.
Proof.
  induction n; simpl; intuition.
Qed.


Lemma isOddisEvenOk : forall n,
  (if isOdd n then odd n else even n) /\
  (if isEven n then even n else odd n).
Proof.
  induction n.
  simpl; intuition.
  simpl.
  intuition.
  destruct (isEven n); intuition.
  destruct (isOdd n); intuition.
Qed.


Lemma isOddOk : forall n,
  if isOdd n then odd n else even n.
Proof.
  intro n.
  assert (Q := isOddisEvenOk n).
  intuition.
Qed.


Lemma isEvenOk : forall n,
  if isEven n then even n else odd n.
Proof.
  intro n.
  assert (Q := isOddisEvenOk n).
  intuition.
Qed.


Lemma beqSBoolEq : forall x:bool,
  beq_nat (if x then 1 else 0) 0 = negb x.
Proof.
  destruct x; reflexivity.
Qed.



Lemma LTeq : forall x y,
  x <= y ->
  exists k, x + k = y.
Proof.
  induction y; intros Lt.
  exists 0; omega.
  compare x (S y).
  intro; subst.
  exists 0; omega.
  intro Neq.
  destruct IHy as [k R].
  omega.
  subst.
  exists (S k).
  omega.
Qed.


Definition aligned4 x y := exists k, y = x + 4 * k.
Definition aligned8 a b := exists k, b = a + 8 * k.


Lemma aligned4Refl : forall x,
  aligned4 x x.
Proof.
  exists 0; omega.
Qed.

Lemma aligned4Trans : forall x y,
  aligned4 x y -> forall z,
    aligned4 y z ->
    aligned4 x z.
Proof.
  intros x y [k1 E1] z [k2 E2].
  subst.
  exists (k1 + k2).
  omega.
Qed.



Lemma aligned8Refl : forall x,
  aligned8 x x.
Proof.
  exists 0; omega.
Qed.


Lemma aligned8Trans : forall x y,
  aligned8 x y -> forall z,
    aligned8 y z ->
    aligned8 x z.
Proof.
  intros x y [k1 E1] z [k2 E2].
  subst.
  exists (k1 + k2).
  omega.
Qed.


Lemma aligned8is4 : forall x y,
  aligned8 x y ->
  aligned4 x y.
Proof.
  destruct 1 as [k E].
  rewrite E; unfold aligned4.
  exists (2 * k); omega.
Qed.



Lemma oddIntro : forall x k,
  x = S (2 * k) ->
  odd x.
Proof.
  intros x k E; subst; constructor.
  apply even_mult_l.
  repeat constructor.
Qed.

  

(* put in nat util *)
Lemma isOddTrue : forall x,
  odd x ->
  isOdd x = true.
Proof.
  intros x O.
  assert (X := isOddOk x).
  destruct (isOdd x).
  reflexivity.
  elimtype False; eapply not_even_and_odd; eauto.
Qed.


(* nat util *)
Lemma isOddFalse : forall x,
  even x ->
  isOdd x = false.
Proof.
  intros x E.
  assert (X := isOddOk x).
  destruct (isOdd x).
  elimtype False; eapply not_even_and_odd; eauto.
  reflexivity.
Qed.


Lemma aligned8LE_L : forall x y,
  aligned8 x y -> forall z,
  aligned8 x z ->
  y <= z ->
  aligned8 y z.
Proof.
  intros x y [k Axy] z [k' Axz] Le.
  subst.
  exists (k' - k); omega.
Qed.


Lemma aligned8LE_R : forall x z,
  aligned8 x z -> forall y,
  aligned8 y z ->
  x <= y ->
  aligned8 x y.
Proof.
  intros x z [k Axz] y [k' Ayz] Le.
  subst.
  exists (k - k'); omega.
Qed.


Lemma aligned8_succ : forall x y,
  aligned8 x y ->
  x <> y ->
  aligned8 (8 + x) y.
Proof.
  intros x y [k E] NE.
  destruct k.
  elf; omega.
  exists k; omega.
Qed.
