Require Import CheneyLemma.


Set Implicit Arguments.


(* code address constants *)
Definition SCANFIELD  := 160.
Definition SCANPOINTER := 169.
Definition SCANCOPIED := 173.
Definition FWDOBJ := 176.
Definition NOTTOSPACEPTR := 184.
Definition SCANNOCOPY := 193.
Definition RETURN := 196.


(* local register conventions *)
Definition rfree := a3.
Definition rtoStart := t5.
Definition rtoEnd := t6.

Notation rtemp := a2.
Notation temp0 := t8.
Notation temp1 := t1.


(* code block definitions *)

Definition fwdObj :=
  (* copy fields from src to free *)
  lw rtemp 0 a0;
  sw rtemp 0 rfree;
  lw rtemp WORDSIZE a0;
  sw rtemp WORDSIZE rfree;
  (* store forwarding pointer in src *)
  sw rfree 0 a0;
  (* return value of free *)
  addiu v0 rfree 0;
  (* increment free *)
  addiu rfree rfree OBJSIZE;
  jr ra.


Notation isNotSpacePtr :=
  (fun rstart rend =>
    andi1 v0 a0;
    bne v0 zero RETURN;;
    sltu v0 a0 rstart;
    bne v0 zero RETURN;;
    addiu v0 a0 1;
    sltu v0 rend v0;
    bne v0 zero RETURN;;
    addiu v0 zero 0;
    jr ra).

Definition isNotToSpacePtr := isNotSpacePtr rtoStart rtoEnd.


Definition scanPointer :=
  addiu rfree t4 0;        (* rfree := t4                *)
  beq v0 zero SCANNOCOPY;; (* if v0 = 0, goto SCANNOCOPY *)
  addiu a0 temp1 0;        (* a0 := temp1                *)
  jal FWDOBJ SCANCOPIED.

Definition scanNoCopy :=
  sw t2 0 temp0;          (* [temp0] := t2 *)
  addiu ra t7 0;          (* ra := t7      *)
  jr ra.

Definition scanCopied :=
  sw v0 0 temp0;          (* [temp0] := v0 *)
  addiu ra t7 0;          (* ra := t7      *)
  jr ra.

Definition scanField :=
  lw temp1 0 a0;           (* temp1 := [a0]    *)
  andi1 rtemp temp1;       (* t := temp1 && 1  *)
  bne rtemp zero RETURN;;  (* if t <> 0 return *)
  addiu temp0 a0 0;        (* temp0 := a0      *)
  lw t2 0 temp1;           (* t2 := [temp1]    *)
  addiu t7 ra 0;           (* t7 := ra         *)
  addiu t4 rfree 0;        (* t4 := rfree      *)
  addiu a0 t2 0;           (* a0 := t2         *)
  jal NOTTOSPACEPTR SCANPOINTER.


(* specs *)

Definition fwdObjPre s :=
  s |||- rget s a0 |->__ ** rget s rfree |->__ ** top.

Definition fwdObjGuar :=
  fun s s' => (forall A x y x' y',
    s |||- rget s a0 |->, (x::y::nil) ** 
      rget s rfree |->, (x'::y'::nil) ** A ->
	s' |||- rget s a0 |->, (rget s rfree::y::nil) ** 
	  rget s rfree |->, (x::y::nil) ** A) /\
	    rget s' rfree = OBJSIZE + rget s rfree /\
	    rget s' v0 = rget s rfree /\
	    regPres (regNotInList (rfree::rtemp::v0::nil)) s s'.


Definition isNotSpacePtrPre : spred := fun s => True.

Notation isNotSpacePtrGuar := (fun rstart rend s s' =>
  sbool (rget s' v0) (notSpacePtrP (rget s a0) (rget s rstart) (rget s rend)) /\
  heapOf s = heapOf s' /\ 
  regPres (fun r => r <> v0) s s').

Definition isNotToSpacePtrGuar := isNotSpacePtrGuar rtoStart rtoEnd.

Definition scanNoCopyPre s :=
  exists v, s |||- rget s temp0 |-> v ** top.

Definition scanNoCopyGuar s s' :=
  (forall v A, s |||- rget s temp0 |-> v ** A ->
    s' |||- rget s temp0 |-> rget s t2 ** A) /\
  rfileEqv s' (rset s ra (rget s t7)).


Definition scanCopiedPre s :=
  exists v, s |||- rget s temp0 |-> v ** top.

Definition scanCopiedGuar s s' :=
  (forall v A, s |||- rget s temp0 |-> v ** A ->
    s' |||- rget s temp0 |-> rget s v0 ** A) /\
  rfileEqv s' (rset s ra (rget s t7)).


Definition scanPointerPre s :=
  if beq_nat (rget s v0) 0 then
    scanNoCopyPre s
  else
    exists v, 
      s |||- rget s temp0 |-> v ** rget s temp1 |->__ ** rget s t4 |->__ ** top.

Definition scanPointerGuar s s' :=
  if beq_nat (rget s v0) 0 then
    scanNoCopyGuar (rset s rfree (rget s t4)) s'
  else
    exists s'',
      fwdObjGuar (rset (rset (rset s rfree (rget s t4)) a0
	(rget s temp1)) ra SCANCOPIED) s'' /\ 
      scanCopiedGuar s'' s'.

Definition scanFieldStOk phi W M F A s :=
  (set_card W) <= (set_card F) /\
  set_disj (W |_| M) (setRange (rget s rtoStart) (rget s rtoEnd)) /\
  set_eqv F (rangeObjs (rget s rfree) (rget s rtoEnd)) /\
  (forall x y, phi x = Some y -> 
    ~notSpacePtrP y (rget s rtoStart) (rget s rtoEnd)) /\
  rget s rtoStart <= rget s rfree <= rget s rtoEnd /\
  s |||- mapHp M phi ** objHp (W |_| M) W **
    (foralls x in F, x |->__) ** A.


Definition scanFieldPre s :=
  exists phi, exists W, exists M, exists F,
    scanFieldStOk phi W M F (okFieldAt (W |_| M) (rget s a0) ** top) s.

Definition scanFieldGuar s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F (okFieldAt (W |_| M) (rget s a0) ** A) s ->
    (exists phi', exists W', exists M', exists F', exists G',
      set_eqv W (M' |_| W') /\
      set_eqv F (G' |_| F') /\
      iso_set phi' M' G' /\
      (forall x, set_In x G' -> Heap.get (heapOf s') (rget s a0) = Some x) /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      (forall x, set_In x W' -> s' |||- objCopied x s x ** top) /\
      scanFieldStOk (map_union phi phi') W' (M |_| M') F'
      (fieldRespMap (map_union phi phi') s (rget s a0) **
	(foralls x in G', objCopiedPhi phi' s x) ** A) s')) /\
    regPres 
    (regNotInList (rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s'.


(* code heap spec *)

Definition Pscanf : chspec := fun l => 
  if beq_nat l RETURN then
    SomeT (bspeci returnPre returnGuar)
  else if beq_nat l SCANPOINTER then
    SomeT (bspeci scanPointerPre scanPointerGuar)
  else if beq_nat l NOTTOSPACEPTR then
    SomeT (bspeci isNotSpacePtrPre isNotToSpacePtrGuar)
  else if beq_nat l SCANCOPIED then
    SomeT (bspeci scanCopiedPre scanCopiedGuar)
  else if beq_nat l SCANNOCOPY then
    SomeT (bspeci scanNoCopyPre scanNoCopyGuar)
  else if beq_nat l FWDOBJ then
    SomeT (bspeci fwdObjPre fwdObjGuar)
  else if beq_nat l SCANFIELD then
    SomeT (bspeci scanFieldPre scanFieldGuar)
  else
    NoneT _.



Ltac PscanfSimpl := 
  unfold Pscanf; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(* lemmas related to the scanField invariant *)


Lemma phiWEq : forall phi W M F P s,
  scanFieldStOk phi W M F P s ->
  forall x, set_In x W -> phi x = None.
Proof.
  unfold scanFieldStOk.
  intuition.
  assert (~set_In x M).
  intro IM.
  hyr H5 (sconjImplL H5 _ (mapHpElim1 IM (phi:=phi))).
  lle H5.
  sconjLift2 H5.
  hyr H5 (sconjImplL H5 _ (objHpElim H3 (s1 := W |_| M))).
  sconjLift2 H5.
  unfold okObjAt in H5; unfold okFieldAt in H5.
  lle H5.
  refine (sconjifierL (mapHpElim2 H7 (phi:=phi)) H5).
Qed.


Lemma scanFieldPtrW : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall x, set_In x W ->
    ptrP x.
Proof.
  unfold scanFieldStOk.
  intros.

  assert (hget s x <> None).
  intuition.
  sconjLift2 H7.
  hyr H7 (sconjImplL H7 _ (objHpElim H0 (s1 := _))).
  unfold okObjAt in H7; unfold okFieldAt in H7; lle H7.
  assert (X := contGetEqL H7).
  unfold hget in H1.
  rewrite H1 in X; discriminate X.
  unfold hget in H1; unfold Heap.get in H1.
  assert (U := isPtrOk x).
  destruct (isPtr x).
  assumption.
  elf; apply H1; reflexivity.
Qed.


Lemma scanFieldGuarPresRa : forall s s',
  scanFieldPre s -> 
  scanFieldGuar s s' -> 
  rget s ra = rget s' ra.
Proof.
  unfold scanFieldPre; unfold scanFieldGuar.
  intros s s' _ [_ Rp].
  rewrite (regPresEq Rp).
  reflexivity.
  regNotInListSolver.
Qed.


Lemma scanFieldStOkTop : forall phi W M F P P' s,
  scanFieldStOk phi W M F (P ** P') s ->
  scanFieldStOk phi W M F (P ** top) s.
Proof.
  unfold scanFieldStOk.
  intuition; scm.
Qed.


Lemma scanFieldStOkImpl : forall phi W M F P s ,
  scanFieldStOk phi W M F P s ->
  forall (P':hpred), (forall h, h ||- P -> h ||- P') ->
  scanFieldStOk phi W M F P' s.
Proof.
  unfold scanFieldStOk.
  intuition; scm.
Qed.


Lemma scanFieldStOkWMDisj : forall phi W M F P s,
  scanFieldStOk phi W M F P s ->
  set_disj W M.
Proof.
  destruct s; unfold scanFieldStOk; simpl.
  unfold mapHp; unfold objHp.
  intuition.
  clear H0 H H1 H2 H4 H6.
  lle H5.
  clear H1.
  apply set_disjSymm.
  apply (fasFasDisj H5); clear h H5 P; intros x h H.
  lleim.
  hpunfoldIn H; lleim.
Qed.


Lemma scanFieldStOkFGDisj : forall phi W M F B G A s,
  scanFieldStOk phi W M F (fas B G ** A) s -> 
  (forall h x, h ||- B x -> h ||- x |->_ ** top) ->
  set_disj F G.
Proof.
  unfold scanFieldStOk.
  intros phi W M F B G A.
  destruct s.
  intros [_ [_ [_ [_ [_ H]]]]] Imp.
  eapply fasFasDisj.
  assert (Q := H).
  sconjLift3 Q; apply (sconjImplR Q); clear Q; intros h' Q.
  sconjLift3 Q; apply (sconjImplR Q); clear h' Q; intros h' Q.
  apply Q.
  simpl; intros h' x H'; apply (sconjImplR H'); intros; lli.
  assumption.
Qed.
Implicit Arguments scanFieldStOkFGDisj [phi W M F B G A s].


Lemma scanFieldStOkMorphHalf : forall s1 s1' s2 s2' s3 s3' u A s s',
  set_eqv s1 s1' ->
  set_eqv s2 s2' ->
  set_eqv s3 s3' ->
  state_eqv s s' ->
  scanFieldStOk u s1 s2 s3 A s ->
  scanFieldStOk u s1' s2' s3' A s'.
Proof.
  unfold scanFieldStOk.
  intros until s'; intros E1 E2 E3 Seqv.
  repeat (rewrite <- (rgetEqv Seqv)).
  intuition.

  rewrite <- (set_cardMorph E1).
  rewrite <- (set_cardMorph E3).
  assumption.

  setoid_rewrite <- E1.
  setoid_rewrite <- E2.
  assumption.

  setoid_rewrite <- E3.
  assumption.
  
  clear H H1 H2 H4 H6.
  apply (heapEqvTrue (stateHeapEqv (state_eqvSymm Seqv))).
  apply (sconjImpl H5); clear H5; intros h H.
  unfold hpredTrue.
  setoid_rewrite <- E2.
  assumption.

  scm.
  unfold hpredTrue.
  setoid_rewrite <- E1.
  setoid_rewrite <- E2.
  assumption.
  unfold hpredTrue.
  setoid_rewrite <- E3.
  assumption.

Qed.

Add Morphism scanFieldStOk
  with signature eq ==> set_eqv ==> set_eqv ==> set_eqv ==> eq ==> state_eqv ==> iff
  as scanFieldStOkMorphism.
  intros u s1 s1' E1 s2 s2' E2 s3 s3' E3 A s s' Seqv; constructor; intros;
    eapply scanFieldStOkMorphHalf; eauto; try (apply set_eqvSymm; auto).
  apply state_eqvSymm; auto.
Qed.


Lemma scanFieldStOkPhiEqv : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall phi', mapEqv phi phi' ->
    scanFieldStOk phi' W M F A s.
Proof.
  unfold scanFieldStOk.
  intuition.
  assert (Q := H2 x y).
  rewrite (H3 x) in Q; auto.
  clear H H1 H2 H4 H6; rename H3 into E; rename H5 into H.
  apply (sconjImplL H); clear s H; intros h H.
  exact (mapHpPhiEqv H E).
Qed.


Notation heap_eqv := Heap.heap_eqv.

Lemma scanFieldStOkExtract : forall phi W M F A B s,
  scanFieldStOk phi W M F (A ** B) s ->
  exists h, scanFieldStOk phi W M F (A ** heap_eqv h) s /\ h ||- B.
Proof.
  unfold scanFieldStOk; intuition.
  sconjLift5of5 H5.
  destruct (sconjExtract H5) as [h [H' HB]].
  exists h.
  clear H5.
  intuition.
  clear H0 H H1 H2 H4 H6 HB.
  hyr H' (sconjComm H'); lle H'.
  assumption.
Qed.


Lemma scanPhiDom : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall x, ~set_In x M -> 
    phi x = None.
Proof.
  unfold scanFieldStOk.
  intros phi W M F A s H x NI.
  intuition.
  assert (s |||- (!(phi x = None) ** top) ** top).
  apply (sconjImpl H5); clear H; intros h H.
  hyr H (mapHpElim2 NI H).
  lli.
  lli.
  lle H3.
  assumption.
Qed.


(* block verification *)

Lemma scanNoCopyOk : iseqOkS Pscanf scanNoCopyPre scanNoCopyGuar scanNoCopy.
Proof.
  unfold scanNoCopyPre; unfold scanNoCopyGuar; unfold scanNoCopy.
  iseqWeakSolver PscanfSimpl.
  intros s [v Hp].
  msteps Hp.
  constructor.
  trivial.
  intros; subst; stSimpl.
  constructor.
  intros v0 A Hp0.
  simpl.
  apply (contUpdateSt Hp0).
  intro r; destruct r; stSimpl; reflexivity.
Qed.


Lemma scanCopiedOk : iseqOkS Pscanf scanCopiedPre scanCopiedGuar scanCopied.
Proof.
  unfold scanCopiedPre; unfold scanCopiedGuar; unfold scanCopied.
  iseqWeakSolver PscanfSimpl.
  intros s [v Hp].
  msteps Hp.
  constructor.
  trivial.
  intros; subst; stSimpl; simpl.
  constructor.
  intros v0 A Hp'.
  apply (contUpdateSt Hp').
  intro r; destruct r; stSimpl; reflexivity.
Qed.


Lemma scanPointerOk : iseqOkS Pscanf scanPointerPre scanPointerGuar scanPointer.
Proof.
  unfold scanPointerPre; unfold scanPointerGuar; unfold scanPointer.
  iseqWeakSolver PscanfSimpl.

  unfold fwdObjGuar.
  intros s s' _ [_ [_ [_ Req]]].
  apply Req.
  regNotInListSolver.

  intros s Hp.
  cbv beta; bsteps.

  rewrite Bt in Hp; rewrite beq_true in Hp.
  rewrite Bt; rewrite beq_true.
  intuition.

  rewrite (beq_false Bt) in Hp; rewrite (beq_false Bt).
  destruct Hp as [v Hp].

  constructor; auto.
  unfold fwdObjPre; unfold fwdObjGuar; stSimpl.
  constructor.
  permConj Hp (2::3::1::4); scm.

  intros s' [G [Feq [V0eq Cso]]].
  lle H.
  simpl in G.

  assert (Gx := G (rget s temp0 |-> v ** top) x x0 x1 x2); clear G.
  exists v.
  replace (rget s' temp0) with (rget s temp0).

  assert (X : ((rget s temp1 |-> rget s t4 ** S (S (S (S (rget s temp1)))) |-> x0) **
    (rget s t4 |-> x ** S (S (S (S (rget s t4)))) |-> x0) **
    rget s temp0 |-> v ** top) (Heap.heap_canon (heapOf s'))).
  apply Gx.
  clear Gx Cso V0eq Feq.
  sconjGLift3; lli; assumption.

  clear Hp Gx.
  sconjLift3 X.
  scm.

  rewrite (regPresEq Cso temp0).
  stSimpl; reflexivity.
  regNotInListSolver.

Qed.


Section INSP.

Variable rstart rend : rnum.

Variable rstartOk : rstart <> v0.
Variable rendOk : rend <> v0.

Lemma isNotSpacePtrOk : iseqOkS Pscanf isNotSpacePtrPre 
  (isNotSpacePtrGuar rstart rend) (isNotSpacePtr rstart rend).
Proof.
  iseqWeakSolver PscanfSimpl.

  cbv beta; intros s _.
  unfold returnPre; unfold returnGuar.
  assert (IO := isOddOk (rget s a0)).
  bsteps.

  constructor.
  trivial.

  intros s' E; subst s'; stSimpl.
  destruct (isOdd (rget s a0)).
  simpl.
  constructor.
  left; assumption.
  constructor.
  reflexivity.
  unfold regPres; intros; stSimpl; reflexivity.
  
  elf; auto.
  destruct (isOdd (rget s a0)); [discriminate Bt | clear Bt].
  constructor.
  destruct (blt_nat (rget s a0) (rget s rstart)).
  simpl; trivial.
  destruct (blt_nat (rget s rend) (1 + rget s a0)); simpl; trivial.

  assert (X := blt_natSound (rget s a0) (rget s rstart)).
  assert (Y := blt_natSound (rget s rend) (1 + rget s a0)).
  destruct (blt_nat (rget s a0) (rget s rstart)).

  simpl; intros s' E; subst s'; stSimpl; simpl.
  constructor; [right | constructor; [trivial | idtac]].
  intro Z; hyr Z (setRangeElim Z); omega.
  unfold regPres; intros; stSimpl; reflexivity.

  rreplace (doBtest btne 0 0) with false.
  destruct (blt_nat (rget s rend) (1 + rget s a0));
    simpl; intros s' E; subst s'; stSimpl; simpl; constructor.

  right; intro Z; hyr Z (setRangeElim Z); omega.
  constructor; auto; unfold regPres; intros; stSimpl; reflexivity.

  intros [Z | Z].
  apply (not_even_and_odd (rget s a0)); auto.
  apply Z; clear Z; apply setRangeIntro; omega.
  constructor; auto; unfold regPres; intros; stSimpl; reflexivity.

Qed.

End INSP.


Lemma isNotToSpacePtrOk : iseqOkS Pscanf isNotSpacePtrPre isNotToSpacePtrGuar isNotToSpacePtr.
Proof.
  unfold isNotToSpacePtrGuar; unfold isNotToSpacePtr.
  apply isNotSpacePtrOk; discriminate.
Qed.


Lemma fwdObjOk : iseqOkS Pscanf fwdObjPre fwdObjGuar fwdObj.
Proof.
  unfold fwdObjPre; unfold fwdObjGuar; unfold fwdObj.
  iseqWeakSolver PscanfSimpl.

  intros s H.
  lle H.
  assert (Hp1 := H).
  msteps H.
  sconjLift3 H; msteps H; stSimpl.
  unfold WORDSIZE.
  hyr H (contUpdateSt H x).
  assert (Hp2 := H).
  sconjLift3 H; msteps H; clear H.
  sconjLift4 Hp2; msteps Hp2; stSimpl.
  hyr Hp2 (contUpdateSt Hp2 x0).
  sconjLift3 Hp2; msteps Hp2.
  hyr Hp2 (contUpdateSt Hp2 (rget s rfree)).

  constructor.
  trivial.
  intros s' E; subst s'; stSimpl.
  constructor.
  clear Hp2.
  intros; llei H.
  assert (R1 := contContEq Hp1 H).
  assert (x0 = y).
  sconjLift2 Hp1; sconjLift2 H; exact (contContEq Hp1 H).
  assert (x1 = x').
  sconjLift3 Hp1; sconjLift3 H; exact (contContEq Hp1 H).
  assert (x2 = y').
  sconjLift4 Hp1; sconjLift4 H; exact (contContEq Hp1 H).
  subst.

  sconjLift3 H.
  hyr H (contUpdateSt H x3).
  sconjLift4 H.
  hyr H (contUpdateSt H y).
  sconjLift3 H.
  hyr H (contUpdateSt H (rget s rfree)).
  permConjG (1::4::3::2::5); assumption.

  intuition.

  unfold regPres.
  intros; simpl in H.
  intuition; stSimpl; auto.

Qed.




(* lower level version of scanFieldGuar that 
    should be easier to verify scanField using *)
Definition scanFieldGuar0 s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F (okFieldAt (W |_| M) (rget s a0) ** A) s ->
    ((rget s rfree = rget s' rfree /\
      (forall x, set_In x W -> (s' |||- objCopied x s x ** top)) /\
      scanFieldStOk phi W M F 
      (fieldRespMap phi s (rget s a0) ** A) s') \/
     (8 + rget s rfree = rget s' rfree /\
       exists v, s |||- rget s a0 |-> v ** top /\ set_In v W /\
	 (forall x, set_In x (set_remove v W) ->
           (s' |||- objCopied x s x ** top)) /\
	 scanFieldStOk (map_add phi v (rget s rfree))
         (set_remove v W) (set_add v M) 
         (set_remove (rget s rfree) F) 
         (rget s a0 |-> rget s rfree ** 
	   objCopied v s (rget s rfree) ** A) s'))) /\
  regPres 
  (regNotInList (rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s'.


Lemma fieldRespMapPhiEq : forall h phi s x,
  h ||- fieldRespMap phi s x -> forall phi',
    (forall y, phi y = phi' y) ->
    h ||- fieldRespMap phi' s x.
Proof.
  unfold fieldRespMap.
  intros h phi s x Hp phi' Peq.
  lleim.
  clear Hp.
  destruct H1 as [v [Hp Meq]]; exists v; constructor; [assumption | idtac].
  unfold mapX in Meq |- * .
  destruct (isOdd v); auto; rewrite <- Peq; assumption.
Qed.


Lemma iso_setAdd : forall phi X Y,
  iso_set phi X Y -> forall x,
    ~set_In x X -> forall y,
    ~set_In y Y ->
    iso_set (map_add phi x y) (set_add x X) (set_add y Y).
Proof.
  unfold iso_set; unfold FnMap.map_add.
  intros phi X Y Iso x nIx y nIy.
  destruct Iso as [I1 [I2 [I3 I4]]].
  simpl; constructor.

  intro x0; hyr I1 (I1 x0); destruct I1 as [I1 I1'].
  constructor; intro Q.
  destruct Q.
  subst x0; rewrite beq_true; discriminate.
  destruct (beq_nat x x0); auto; discriminate.
  compare x x0; intro R.
  auto.
  right; rewrite (beq_false R) in Q; auto.
  
  constructor.
  intros x' y' E.
  hyr I2 (I2 x' y').
  compare x x'; intro R.
  subst x'; rewrite beq_true in E; injection E; clear E; intro; subst; auto.
  rewrite (beq_false R) in E; auto.

  constructor.
  intros x' x'' y' E1 E2; hyr I3 (I3 x' x'' y').
  compare x x'; intro R.
  subst x'; rewrite beq_true in E1; injection E1; clear E1; intro; subst y'.
  compare x x''; intro R'.
  auto.
  rewrite (beq_false R') in E2; elf; eauto.
  rewrite (beq_false R) in E1.
  compare x x''; intro R'.
  subst x''; rewrite beq_true in E2.
  injection E2; clear E2; intro; subst y'; elf; eauto.
  rewrite (beq_false R') in E2; eauto.

  intros y' [Z | Z].
  subst y'; exists x; rewrite beq_true; auto.

  destruct (I4 _ Z) as [x' E]; exists x'.
  rewrite beq_false; auto.
  intro; subst x'.
  destruct (I1 x) as [_ U]; apply nIx; apply U; rewrite E; discriminate.

Qed.


(* special case of iso_setAdd *)
Lemma iso_setSing : forall x y,
  iso_set (map_add (empty_map _ _) x y) (set_add x empty_set) (set_add y empty_set).
Proof.
  intros x y.
  apply iso_setAdd.
  apply empty_mapIso.
  simpl; auto.
  simpl; auto.
Qed.


(* scanFieldGuar0 is stronger than scanFieldGuar *)
Lemma scanFieldGuar0Stronger : forall s s',
  scanFieldGuar0 s s' ->
  scanFieldGuar s s'.
Proof.
  intros s s' [X Rp].
  constructor; [idtac | assumption].

  intros phi W M F A H.
  assert (Y := X _ _ _ _ _ H); clear X; rename H into UU.

  assert (RM : set_eqv (M |_| empty_set) M).
  setoid_replace (M |_| empty_set) with (empty_set |_| M).
  apply set_eqvRefl.
  apply set_eqvUnionSymm.


  destruct Y as [H | H].

  (* case: scanned object was not copied *)
  destruct H as [FrEq H].
  exists (fun n:nat => None (A := nat)); 
    exists W; exists empty_set; exists F; exists empty_set.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply empty_mapIso.
  constructor.
  simpl.
  intros; elf; assumption.
  unfold scanFieldStOk in H; unfold scanFieldStOk.
  intuition.
  rewrite FrEq; apply aligned8Refl.
  
  setoid_rewrite RM.
  assumption.
  apply (H3 x y).
  unfold map_union in H4.
  destruct (phi x); assumption.
  assumption.

  clear H0 H H1 H5 H7.
  apply (sconjImpl H6); clear H6; intros h H.
  unfold hpredTrue.
  setoid_rewrite RM.
  rewrite <- hpredTrueEq.
  unfold mapHp; unfold mapHp in H.
  llei H.
  unfold map_union.
  intros x y Z.
  apply (H4 x y).
  destruct (phi x); assumption.
  apply (fasImpl H).
  simpl.
  clear h H; intros x h H.
  lle H.
  existsl x0.
  lli.
  unfold map_union; destruct (phi x); assumption.
  eauto.
  scm.
  unfold hpredTrue; setoid_rewrite RM; assumption.
  apply (fieldRespMapPhiEq Hp).
  unfold map_union; intro y; destruct (phi y); reflexivity.

  lli; auto.

  (* case: scanField did copy an object *)
  destruct H as [FrEq [v [vEq [vIn [Y0 H]]]]].
  exists (map_add (empty_map _ _) v (rget s rfree));
    exists (set_remove v W); exists (set_add v empty_set); 
      exists (set_remove (rget s rfree) F); exists (set_add (rget s rfree) empty_set).
  assert (Pvn := phiWEq UU _ vIn).
  assert (vptr := scanFieldPtrW UU _ vIn).

  constructor.
  simpl; apply set_eqvSymm; apply addRemoveEqv; assumption.
  constructor.
  
  simpl; apply set_eqvSymm; apply addRemoveEqv.
  (* show that F contains (rget s rfree) *)
  unfold scanFieldStOk in UU; intuition.
  setoid_rewrite H1.
  clear RM H2 H.
  destruct (minInObjsRange (rget s rfree) (rget s rtoEnd)).
  assert (set_card W = 0).
  replace (set_card F) with 0 in H0.
  omega.
  assert (set_card (rangeObjs (rget s rfree) (rget s rtoEnd)) = 0).
  rewrite H; reflexivity.
  rewrite <- H2.
  apply set_cardMorph.
  apply set_eqvSymm; assumption.
  assert (Q := cardZeroIsEmpty_set W H2).
  assert (set_In v empty_set).
  setoid_rewrite <- Q; assumption.
  elf; exact H4.
  assumption.

  constructor.
  apply iso_setSing.
  constructor.
  simpl.
  intros x [Xeq | Fa].
  subst x.
  destruct H as [_ X].
  intuition.
  sconjLift4 H4.
  eapply contGetEqL; eauto.
  elf; assumption.

  unfold scanFieldStOk in H |- * .

  assert (MvEq : set_eqv (M |_| (set_add v empty_set)) (set_add v M)).
  setoid_rewrite (set_eqvUnionSymm M (set_add v empty_set)).
  simpl; apply set_eqvRefl.

  intuition.

  rewrite <- FrEq; exists 1; omega.

  setoid_rewrite MvEq; assumption.

  clear H6 H H1 Rp.
  unfold map_union in H3.
  apply (H2 x y); unfold FnMap.map_add.
  unfold FnMap.map_add in H3.
  compare v x; intro R.
  subst v; rewrite Pvn in H3.
  rewrite beq_true in H3; rewrite beq_true.
  assumption.
  rewrite beq_false in H3.
  rewrite beq_false.
  destruct (phi x); assumption.
  assumption.
  auto.
  assumption.

  clear Y0 UU FrEq H0 H H1 H2 H4.
  
  apply (sconjImpl H5); clear H5; intros h H.
  unfold hpredTrue; setoid_rewrite MvEq.
  rewrite <- hpredTrueEq; unfold mapHp in H |- * .
  llei H.
  intros.
  unfold FnMap.map_add in H0; unfold map_union in H0.
  compare x v; intro R.
  rewrite R; apply set_add_intro1; reflexivity.
  apply (H2 x y).
  rewrite beq_false in H0.
  unfold FnMap.map_add.
  rewrite beq_false.
  destruct (phi x); assumption.
  auto.
  auto.

  llMatch.
  lle H; existsl x0; lli.
  unfold FnMap.map_add in H1 |- *; unfold map_union.
  compare x v; intro R.
  subst v; rewrite beq_true; rewrite beq_true in H1.
  rewrite Pvn; assumption.
  rewrite beq_false; rewrite beq_false in H1.
  rewrite H1; reflexivity.
  auto.
  auto.
  auto.
  eauto.

  scm.
  unfold hpredTrue; setoid_rewrite MvEq; assumption.

  unfold fieldRespMap.
  lli.
  exists v.

  constructor.
  assumption.

  unfold FnMap.map_add; unfold map_union; unfold mapX.
  rewrite beq_true.
  rewrite Pvn.
  replace (isOdd v) with false.
  reflexivity.
  assert (Ev := isOddOk v).
  destruct (isOdd v).
  clear Rp; elf.
  eapply not_even_and_odd.
  2:apply Ev.
  apply ptrPeven.
  assert (X := isPtrOk v).
  assumption.
  reflexivity.

  assumption.

  apply fasIntro.
  hyr Hp (unitIR Hp).  
  scm.
  unfold objCopiedPhi; lli.
  unfold FnMap.map_add; rewrite beq_true.
  reflexivity.
  assumption.
  simpl; auto.

Qed.


Lemma scanFieldOk : iseqOkS Pscanf scanFieldPre scanFieldGuar scanField.
Proof.
  assert (iseqOkS Pscanf scanFieldPre scanFieldGuar0 scanField).
  unfold scanFieldPre; unfold scanFieldGuar; unfold scanField.
  iseqWeakSolver PscanfSimpl.

  (* ra is preserved *)
  intros s s' _ [_ [_ C]].
  rewrite C.
  reflexivity.
  discriminate.
  
  intros s [phi [W [M [F [Swf [Disj [Feq [PhiOk [FrG H]]]]]]]]].
  unfold okFieldAt in H.
  lle H; sconjLift4 H.
  msteps H.
  bsteps.

  (* case: field is not a pointer *)
  assert (IOx : odd x).
  assert (IO := isOddOk x); destruct (isOdd x).
  assumption.
  elf; auto.
  rewrite (isOddTrue IOx); clear Bt.
  constructor.

  unfold returnPre; trivial.

  unfold returnGuar; unfold scanFieldGuar0.
  intros s' R; subst s'; stSimpl.
  unfold scanFieldStOk; stSimpl.
  constructor.
  intros phi' W' M' F' A.

  intros [Swf' [Ephi' [Feq' [PhiOk' [FrG' H']]]]].

  left; stSimpl; intuition.

  (* proving that all objects in W' are copied *)
  rename x0 into x'; rename H4 into I; clear H0.
  unfold objCopied; unfold fieldCopied.
  sconjLift2 H'.
  assert (s |||- (x' |->__ ** top) ** top).
  assert (X := H').
  scm.
  hyr Hp (objHpElim I Hp).
  scm.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  lli; eauto.
  llei H0.
  scm.
  eauto.
  sconjLift2 H0; scm.

  (* showing that the heap matches up correctly *)
  assert (X := H').
  unfold okFieldAt in X; lle X.
  assert (x0 = x).
  sconjLift4 X.
  apply (contContEq X H).
  subst.
  unfold fieldRespMap.
  lli; eauto.
  exists x; constructor.
  scm.

  assert (Q := isOddOk x).
  unfold mapX.
  destruct (isOdd x).
  reflexivity.
  elf; eapply not_even_and_odd; eauto.

  (* showing the registers are preserved *)
  unfold regPres.
  simpl.
  intuition.
  stSimpl.
  reflexivity.

  (* case: field is a pointer *)
  assert (IEx : even x).
  assert (IO := isOddOk x); destruct (isOdd x).
  discriminate Bt.
  assumption.
  rewrite (isOddFalse IEx); clear Bt.

  unfold okFieldVal in H10.
  destruct H10.
  unfold atomP in H0; elf; eapply not_even_and_odd; eauto.

  destruct (set_union_elim H0); clear H0.

  (* subcase: x in W *)
  sconjLift3 H.
  assert (Q := objHpElimField _ H H1).
  unfold okFieldAt in Q; lle Q.
  msteps Q.

  assert (NSP0 : notSpacePtrP x0 (rget s rtoStart) (rget s rtoEnd)).
  destruct H5.
  left; assumption.
  right.
  intro I.
  apply (set_disj_elim Disj _ H0 I).

  apply callChain.
  constructor.
  unfold isNotSpacePtrPre; trivial.

  (* guarantee *)
  unfold isNotToSpacePtrGuar; stSimpl; simpl.
  intros s' [Bok [Heq Cok]].
  unfold scanPointerPre; unfold scanNoCopyPre.
  replace (rget s' temp0) with (rget s a0);
    [idtac | rewrite <- Cok; [stSimpl; reflexivity | discriminate]].

  rewrite <- Heq.

  assert (NE : rget s' v0 <> 0).
  intro X; rewrite X in Bok; simpl in Bok.
  auto.

  rewrite (beq_false NE); clear Bok.
  constructor.
  exists x.
  sconjLift2 H.
  replace (rget s' temp1) with x; 
    [idtac | rewrite <- Cok; [reflexivity | discriminate]].
  replace (rget s' t4) with (rget s rfree);
        [idtac | rewrite <- Cok; [reflexivity | discriminate]].
  apply (sconjImplR H); clear H; intros h0 H; clear Cok.

  assert (h0 ||- (x |->__ ** top) ** (rget s rfree |->__ ** top) ** top).
  permConj H (1::3::2::4).
  scm.
  assert (X := objHpElim H1 Hp).
  scm; unfold okObjAt in Hp0; unfold okFieldAt in Hp0; lleim.

  assert (frIF : set_In (rget s rfree) F).

  assert (FNZ : set_card F <> 0).
  assert (WNZ : set_card W <> 0).
  intro Z.
  assert (T := cardZeroIsEmpty_set _ Z).
  exact (empty_set_elim (set_eqv_elim1 T _ H1)).
  omega.
  destruct (minInObjsRange (rget s rfree) (rget s rtoEnd)) as [EE | EE].
  rewrite EE in Feq.
  rewrite (set_cardMorph Feq) in FNZ.
  elf; exact (FNZ (empty_setCard)).
  exact (set_eqv_elim2 Feq _ EE).
  assert (HF := fasElim frIF Hp).
  scm.

  clear H Disj.
  hyr H0 (sconjAssocL H0).
  sconjLift3 H0; hyr H0 (sconjAssocL H0).
  sconjLift3 H0.
  scm.

  (* guarantee *)
  unfold isNotToSpacePtrGuar; unfold scanPointerGuar.
  rewrite (beq_false NE).
  intros s'' HB.
  destruct HB as [s''a [Guar1 Guar2]].
  constructor.

  (* show the heap is okay *)
  intros phi' W' M' F' A H'.
  destruct H' as [WFle' [WMdisj' [F'eq [NSP' [[RTle RTle'] H']]]]].
  destruct Guar2 as [Guar2 Req''].
  right.
  repeat (rewrite Req''; stSimpl).
  destruct Guar1 as [Guar1 Req'].
  generalize dependent Req'; stSimpl.
  repeat (rewrite (regPresEq Cok); [stSimpl | discriminate]).
  intros [FrEq [v0Eq Req']]; rewrite FrEq.
  constructor.
  reflexivity.
  exists x; constructor.
  sconjLift2 H; scm.
  unfold okFieldAt in H'; lle H'.
  sconjLift2 H; sconjLift4 H'; assert (E := contContEq H' H); subst x1.

  assert (Ix : set_In x W'); [idtac | clear H12].
  destruct H12 as [xOdd | Ix0].
  elf; apply (not_even_and_odd x); assumption.
  destruct (set_union_elim Ix0) as [Ix | Ix]; clear Ix0; [auto | elf].
  sconjLift2 H'; hyr H' (sconjImplL H' _ (mapHpElim1 Ix (phi:=phi'))).
  lle H'.
  assert (R := contContEq H' Q).
  rewrite R in H6; rewrite R in H'; clear R x1.
  exact (NSP' _ _ H6 NSP0).

  constructor.
  assumption.
  generalize dependent Guar1; generalize dependent Guar2; stSimpl.
  repeat (rewrite <- Req'; [stSimpl | regNotInListSolver]).
  rewrite v0Eq.
  repeat (rewrite (regPresEq Cok); [stSimpl | discriminate]).
  intros.
  rewrite <- Heq in Guar1.

  destruct (minInObjsRange (rget s rfree) (rget s rtoEnd)).

  (* subsubcase: free -- end is empty *)
  elf.
  assert (set_card F' <> 0).
  assert (set_card W' <> 0).
  intro Z.
  assert (T := cardZeroIsEmpty_set _ Z).
  exact (empty_set_elim (set_eqv_elim1 T _ Ix)).
  omega.
  rewrite H0 in F'eq.
  rewrite (set_cardMorph F'eq) in H2.
  elf; exact (H2 (empty_setCard)).

  (* subsubcase: free -- end contains free *)
  assert (If : set_In (rget s rfree) F').
  rewrite F'eq; assumption.

  assert (Hp : s |||- rget s a0 |-> x **
    mapHp M' phi' ** (okObjAt (W' |_| M') x ** objHp (W' |_| M') (set_remove x W')) **
    (rget s rfree |->__ ** (foralls x in set_remove (rget s rfree) F', x |->__)) ** A).
  scm.
  hyr Hp (objHpElim Ix Hp); scm.
  hyr Hp (fasElim If Hp); scm.
  hpunfoldIn Hp; lle Hp; rename x3 into f0; rename x4 into f1.

  assert (Hp' : s |||- 
    rget s a0 |-> x ** (mapHp M' phi' ** objHp (W' |_| M') (set_remove x W') **
      (foralls x in set_remove (rget s rfree) F', x |->__) ** A) **
    x |->, (x1::x2::nil) ** rget s rfree |-> f0 ** 4 + rget s rfree |-> f1).
  lli.
  permConjG (1::2::6::7::3::8::9::4::5); assumption.

  sconjLift2 Hp'; clear Hp; destruct (sconjExtract Hp') as [h [Hp HHp]]; clear Hp'.
  sconjLift4of4 Hp; sconjLift4of4 Hp.
  hyr Guar1 (Guar1 _ _ _ _ _ Hp).
  sconjLift4of4 Guar1; hyr Guar2 (Guar2 _ _ Guar1); clear Guar1.
  constructor.

  intros y Iy.
  assert (ZHp : h ||- (y |->__ ** top) ** top).
  sconjLift2 HHp; scm; hyr Hp0 (objHpElim Iy Hp0); hpunfoldIn Hp0.
  lle Hp0; lli.
  instantiate (1 := x4).
  instantiate (1 := x3).
  scm.
  lle ZHp.
  sconjLift3 Hp; hyr Hp (sconjInsert Hp ZHp); sconjLift4of4 Guar2; 
    hyr Guar2 (sconjInsert Guar2 ZHp).
  lle Hp; lle Guar2; hpunfold; lli.
  instantiate (1 := x3); scm.
  instantiate (1 := x4); scm.
  sconjLift2 Hp; scm.
  unfold scanFieldStOk.
  constructor.

  rewrite <- (set_card_remove _ _ Ix) in WFle'.
  rewrite <- (set_card_remove _ _ If) in WFle'.
  omega.

  repeat (rewrite Req''; stSimpl).
  rewrite FrEq.
  repeat (rewrite (regPresEq Req'); [stSimpl | regNotInListSolver]).
  repeat (rewrite (regPresEq Cok); [stSimpl | discriminate]).

  constructor.
  setoid_rewrite <- (unionRemoveAddEqv _ _ M' Ix).
  assumption.

  constructor.
  setoid_rewrite F'eq.
  apply rangeRemoveEqv.

  constructor.
  intros y z Peq; unfold FnMap.map_add in Peq.
  compare x y; intro R.
  subst y; rewrite beq_true in Peq.
  injection Peq; intros; subst.
  intros X; destruct X.
  assert (HH : s |||- (!(ptrP (rget s rfree)) ** top) ** top).
  sconjLift2 Hp; scm; lle Hp; scm; lli; apply (contElim1L Hp).
  apply not_even_and_odd with (rget s rfree).
  apply ptrPeven; lle HH; assumption.
  auto.
  apply H2; clear H2.
  destruct (rangeObjsElim H0) as [Hz [k Hz']].
  eapply setRangeIntro.
  omega.

  rewrite beq_false in Peq.
  apply (NSP' y z); assumption.
  assumption.

  constructor.
  destruct (rangeObjsElim H0) as [D [k D']].
  unfold OBJSIZE; omega.

  rename Guar2 into Hp3; sconjLift4of4 Hp3; hyr Hp3 (sconjInsert Hp3 HHp); lle Hp3;
    rename h into h0; rename Hp into Hp0; clear HHp Cok PhiOk.
  aspConj Hp3 ([|[|1, [|6, 7|]|], [|2, [|3, [|5, [|[|8, 9|], 4|]|]|]|]|]).
  scm.
  apply mapHpIntro with x2; assumption.
  unfold hpredTrue; setoid_rewrite <- (unionRemoveAddEqv _ _ M' Ix); assumption.
  hpunfold; simpl in Hp; llei Hp0.
  instantiate (1 := x1); scm.
  eauto.
  sconjLift2 Hp0; scm.

  unfold regPres; simpl; intuition.
  destruct Guar2 as [_ Req'']; rewrite Req''; clear Req''; stSimpl.
  destruct Guar1 as [_ [E1 [E2 Req']]].

  destruct (regCompare r ra).
  subst r; stSimpl.
  rewrite (regPresEq Req'); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Cok); [stSimpl; reflexivity | discriminate].
  stSimpl.
  rewrite (regPresEq Req'); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Cok); [stSimpl; reflexivity | assumption].

  (* sub case: x in M *)
  sconjLift2 H.
  assert (Mok := sconjImplL H _ (mapHpElim1 H1 (phi:=phi))).
  lle Mok.
  msteps Mok.
  apply callChain.
  unfold isNotSpacePtrPre; unfold isNotToSpacePtrGuar; stSimpl.
  constructor.
  trivial.
  intros s' [Spv0 [Heq Cso]].
  unfold scanPointerPre.

  assert (Ez : rget s' v0 = 0).
  destruct (rget s' v0).
  reflexivity.
  simpl in Spv0.
  elf; assert (X := PhiOk _ _ H5); auto.

  rewrite Ez; rewrite beq_true.
  rewrite Ez in Spv0; simpl in Spv0.

  unfold scanNoCopyPre.
  rewrite <- Heq.
  rewrite (regPresEq Cso temp0); [stSimpl | discriminate].
  constructor.
  exists x; sconjLift2 H; scm.

  (* post condition *)
  unfold scanPointerGuar.
  rewrite Ez; rewrite beq_true.
  unfold scanNoCopyGuar; unfold fwdObjGuar; unfold scanFieldGuar; stSimpl.
  rewrite <- Heq.
  repeat (rewrite (regPresEq Cso); [stSimpl | discriminate]).
  intros s'' [Guar Req].
  unfold scanFieldGuar0.
  constructor.

  intros phi' W' M' F' A Hp'.
  left.

  constructor.
  rewrite Req; stSimpl; reflexivity.

  constructor.
  intros x' I'.
  unfold scanFieldStOk in Hp'; intuition.
  clear H H0 H2 H3 H6 H4 H7 H9 H11; rename H10 into H.
  unfold okFieldAt in H; lle H; sconjLift4 H.
  assert (H'' : s |||- rget s a0 |-> x1 ** 
    (((lexists v1, lexists v2, x' |->, (v1::v2::nil)) ** top) ** top)).
  apply (sconjImplR H); clear H; intros h H.
  sconjLift2 H; apply (sconjImpl H); clear h H; intros h H.
  hyr H (objHpElim I' H).
  apply (sconjImpl H); clear h H.
  unfold okObjAt; unfold okFieldAt; intros h H; lle H.
  simpl; lli; eauto.
  intros; lli.
  lli.
  lle H''.
  assert (H2 := Guar _ _ H''); clear H Cso Mok PhiOk Feq Disj Guar.
  unfold objCopied; unfold fieldCopied.
  lli.
  instantiate (1 := x2); sconjLift2 H''; scm.
  instantiate (1 := x3); sconjLift3 H2; sconjLift3 H2; scm.
  sconjLift3 H''; scm.

  unfold scanFieldStOk in Hp' |- * .
  repeat (rewrite Req); clear Req; stSimpl.
  repeat (rewrite (regPresEq Cso); [stSimpl | discriminate]).
  clear Cso s' Heq Ez.
  intuition.
  unfold okFieldAt in H10; lle H10.
  sconjLift2 Mok; sconjLift4 H10; assert (R := contContEq H10 Mok).
  rewrite R in H21; rewrite R in H10; clear x1 R.

  assert (I' : set_In x M'); [idtac | clear H21].
  destruct H21.
  elf; unfold atomP in H8; eapply not_even_and_odd; eauto.
  destruct (set_union_elim H8) as [I' | I']; clear H8; auto; elf.
  sconjLift3 H10.
  assert (s |||- (okObjAt (W' |_| M') x ** top) ** top).
  apply (sconjImpl H10); clear H10 Mok H; intros h H.
  hyr H (objHpElim I' H).
  apply (sconjImplR H); intros; lli.
  lli.
  unfold okObjAt in H8; unfold okFieldAt in H8; lle H8.
  sconjLift2 Mok.
  assert (R := contContEq H8 Mok); subst.
  apply (PhiOk _ _ H5).
  destruct H16.
  left; assumption.
  right.
  intro Ni; exact (set_disj_elim H6 _ H12 Ni).

  assert (X := Guar _ _ H10); clear Guar.
  clear H0 H2 H3 H6 H4 H7 H9 H11 Swf Disj Feq PhiOk H1 H.
  permConj X (2::3::4::1::5); scm.
  unfold fieldRespMap.
  existsl x0; lli.
  exists x.
  constructor.
  scm.
  unfold mapX.
  assert (O := isOddOk x).
  destruct (isOdd x).
  elf; eapply not_even_and_odd; eauto.
  clear O.
  sconjLift2 H10.
  assert (s |||- ((lexists v, !(phi' x = Some v) ** x |-> v ** top) ** top)).
  scm; exact (mapHpElim1 I' Hp0).
  lle H.
  sconjLift2 Mok.
  rewrite (contContEq Mok H); assumption.
  assumption.

  (* regPres solving *)
  unfold regPres; simpl; intuition.
  rewrite Req.
  destruct (regCompare r ra).
  subst r; stSimpl.
  reflexivity.
  stSimpl.
  rewrite (regPresEq Cso); stSimpl.
  reflexivity.
  assumption.

  (* now we use the lemma we proved before to show that the implementation spec
     matches the abstract spec *)
  apply (iseqOkWeak _ _ _ _ H); clear H.
  intros s H.
  constructor.
  assumption.
  intros s' H'.
  exact (scanFieldGuar0Stronger H').

Qed.
