Require Import CAP.
Require Import StateTheory.


Section Soundness.

Variable bspec : Type.
Notation chstpred := (chstpred bspec).
Variable interp : bspec -> chstpred.

Notation "a ==> b" :=
  (forall p s, (a:chstpred) p s -> (b:chstpred) p s) (at level 13).

Notation iseqOk := (iseqOk interp).

Lemma iseqOkWeaken : forall a i,
  iseqOk a i ->
  forall a', (a' ==> a) -> iseqOk a' i.
induction 1.

(* j *)
constructor.
intros; apply H.
auto.

(* jal *)
constructor.
intros; apply H.
auto.

(* jr *)
constructor.
intros; apply H.
auto.

(* br *)
econstructor.
apply (IHiseqOk a'); auto.
auto.

(* seq *)
econstructor.
apply (IHiseqOk a'); auto.
intros; apply H0; auto.

Qed.


Lemma cdHeapOkDomSame : forall P C P' f sp,
  cdHeapOk interp P C P' -> P' f = SomeT sp ->
  C f <> None.
induction 1.
intro Eq.
assert (Z := H f sp Eq); clear H.
destruct (C f); [discriminate | auto].
destruct (P1' f).
injection 1; intro; subst; clear H3.
destruct (C1 f).
discriminate.
elimtype False.
apply IHcdHeapOk1.
reflexivity.
reflexivity.

destruct (P2' f).
injection 1; intro; subst; clear H3.
destruct (C1 f).
discriminate.
apply IHcdHeapOk2.
reflexivity.
discriminate 1.
Qed.


Lemma jumpOk : forall P C P' f,
  cdHeapOk interp P C P' -> forall sp, P' f = SomeT sp ->
  optPred (C f) (iseqOk (lift (interp sp) P)).
intros P C P' f H.
induction H; intros.
apply H; auto.
assert (Z : (exists i, P1' f = SomeT i) \/ (exists i, P2' f = SomeT i)).
destruct (P1' f).
left; econstructor; auto.
right; destruct (P2' f).
econstructor; auto.
discriminate H3.
destruct Z.

destruct H4 as [sp' sp'Eq].
clear IHcdHeapOk2.
rewrite sp'Eq in H3.
injection H3; intro; subst; clear H3.
assert (Z := IHcdHeapOk1 _ sp'Eq); clear IHcdHeapOk1.
destruct (C1 f).
eapply iseqOkWeaken.
eauto.
unfold lift.
clear sp'Eq H H0 H1 P1' P2' Z C1 C2.
intros p s [A B].
constructor.
clear B.
intros q r.
assert (Z := A q r); clear A.
destruct (P1 q).
assumption.
intro M; discriminate M.
assumption.
contradiction.

destruct H4 as [sp' sp'Eq].
clear IHcdHeapOk1.
rewrite sp'Eq in H3.

assert (Z := cdHeapOkDomSame _ _ _ _ _ H0 sp'Eq).
assert (QQ : C1 f = None).
assert (~ (C1 f <> None /\ C2 f <> None)).
intro; elimtype False; apply H1.
econstructor; eauto.
destruct (C1 f).
elimtype False.
apply H4.
constructor.
discriminate.
assumption.
reflexivity.
rewrite QQ.
destruct (C2 f).
assert (sp = sp').
assert (Y := fun sp => cdHeapOkDomSame _ _ _ f sp H).
destruct (P1' f).
elimtype False.
exact (Y _ H3 QQ).
injection H3; auto.
subst.
assert (Q := IHcdHeapOk2 _ sp'Eq); clear IHcdHeapOk2.
eapply iseqOkWeaken.
eauto.
unfold lift.
clear QQ H3 sp'Eq H H0 H1 P1' P2' Z C1 C2.
intros p s [A B].
constructor.
clear B.
intros q r.
assert (Z := A q r); clear A.
assert (Y := H2 q); clear H2.
destruct (P2 q).
injection 1; clear H; intro; subst.
apply Z; clear Z.
destruct (P1 q).
apply Y; discriminate.
reflexivity.
intro H; discriminate.
assumption.
apply Z; reflexivity.
Qed.


Lemma liftImpl : forall a P s,
  interp a P s ->
  lift (interp a) P P s.
unfold lift.
intuition.
Qed.


Notation progOk := (progOk interp).

Theorem capSoundness : forall s,
  progOk s -> optPred (step s) progOk.
destruct s as [C s I].
inversion_clear 1 as [P X Y Z a Cok Aok Iok].

inversion Iok as
  [ Z f fOk | AA f fret fOk | AA r rOk |
    AA a' bt rs rt f i iOk bOk | AA a' c i iOk stepOk];
  clear Iok; subst.

(* j *)
simpl.
assert (Q := fOk _ _ Aok); clear fOk.
assert (Z := jumpOk _ _ _ f Cok).
destruct (P f).
destruct (C f).
assert (Y : iseqOk (lift (interp b) P) i).
apply Z.
reflexivity.
clear Z.

econstructor.
apply Cok.
apply (liftImpl _ _ _ Q); assumption.
assumption.
eauto.
contradiction.

(* jal *)
simpl.
assert (Q := fOk _ _ Aok); clear fOk Aok.
assert (Z := jumpOk _ _ _ f Cok).
destruct (P f).
destruct (C f).
assert (Y : iseqOk (lift (interp b) P) i).
apply Z.
reflexivity.
clear Z.

econstructor.
apply Cok.
apply (liftImpl _ _ _ Q); assumption.
assumption.
eauto.
contradiction.

(* jr *)
simpl.
assert (Q := rOk _ _ Aok); clear rOk Aok.
assert (Z := jumpOk _ _ _ (rget s r) Cok).
destruct (P (rget s r)).
destruct (C (rget s r)).
assert (Y : iseqOk (lift (interp b) P) i).
apply Z.
reflexivity.
clear Z.

econstructor.
apply Cok.
apply (liftImpl _ _ _ Q); assumption.
assumption.
eauto.
contradiction.

(* br *)
simpl.
destruct (bOk _ _ Aok) as [Q R]; clear bOk Aok.
assert (BTok := doBtestSound bt (rget s rs) (rget s rt)).
destruct (doBtest bt (rget s rs) (rget s rt)).

(* br - jump case *)
assert (H := R BTok); clear R BTok.
assert (Z := jumpOk _ _ _ f Cok).
destruct (P f).
destruct (C f).
assert (Y : iseqOk (lift (interp b) P) i0).
apply Z.
reflexivity.

clear Z.
econstructor.
apply Cok.
apply (liftImpl _ _ _ H); assumption.
assumption.
eauto.
contradiction.

(* br - fall through *)
clear R.
assert (X := Q BTok); clear Q BTok.
econstructor.
apply Cok.
apply X.
assumption.

(* seq *)
assert (Q := stepOk _ _ Aok); clear stepOk Aok.
simpl.

destruct (cstep c s).
destruct c; econstructor; eauto.
assumption.

Qed.


Theorem stateOkSound : predSound progOk.
unfold predSound.
intros s sOk.
assert (sOk' := capSoundness s sOk); simpl in sOk'.
destruct (step s) as [s' | ].
econstructor; intuition.
intuition.
Qed.


End Soundness.
