Require Export SCAP.
Require Import EqNat.
Require Import StateTheory.


(* weakest precondition.  sort of. *)
Inductive iseqOkSWP (P:chspec) : spred -> guar -> iseq -> Prop :=
  | retOkWP : forall p g,
    p = (fun s => True) ->
    g = (fun s s' => s = s') ->
    iseqOkSWP P p g (jr ra)
  | jOkWP : forall (p':spred) (g':guar) f,
    P f = (SomeT (bspeci p' g')) ->
    iseqOkSWP P p' g' (j f)
  | brOkWP : forall bt rs rt f ii p g p' g' p'' g'',
    P f = (SomeT (bspeci p' g')) ->
    iseqOkSWP P p'' g'' ii ->
    p = (fun s => if doBtest bt (rget s rs) (rget s rt) then p' s else p'' s) ->
    g = (fun s s' => if doBtest bt (rget s rs) (rget s rt) then g' s s' else g'' s s') ->
    iseqOkSWP P p g (br bt rs rt f;; ii)
  | seqOkWP : forall c ii p g p' g',
    iseqOkSWP P p' g' ii ->
    p = (fun s => optPred (cstep c s) p') ->
    g = (fun s s'' => optPred (cstep c s) (fun s' => g' s' s'')) ->
    iseqOkSWP P p g (c; ii)
  | callOkWP : forall f fret p g p' g' p'' g'',
    P f = (SomeT (bspeci p' g')) ->
    P fret = (SomeT (bspeci p'' g'')) ->
    (forall s s', p' s -> g' s s' -> rget s ra = rget s' ra) ->
    p = (fun s => p' (rset s ra fret) /\ forall s', g' (rset s ra fret) s' -> p'' s') ->
    g = (fun s s'' => exists s', g' (rset s ra fret) s' /\ g'' s' s'') ->
    iseqOkSWP P p g (jal f fret).


Theorem WpSound : forall P p g ii,
  iseqOkSWP P p g ii ->
  iseqOkS P p g ii.
Proof.
induction 1; subst.

constructor; auto.

econstructor; eauto.

econstructor; eauto.
intro s.
assert (X := doBtestSound bt (rget s rs) (rget s rt)).
destruct (doBtest bt (rget s rs) (rget s rt)).

intro H'.
constructor.
intuition.
intro Xn; elimtype False; exact (btPredNegOk _ _ _ X Xn).

intro H'.
constructor.
intro Xn; elimtype False; exact (btPredNegOk _ _ _ Xn X).
intuition.

econstructor.
eauto.
intros s.
destruct (cstep c s); intuition.

econstructor; eauto.
clear H H0.
intuition.
exists s'.
intuition.

Qed.


Theorem UseWeakIseq : forall P p g ii,
  iseqOkSWP P p g ii -> forall (p':spred) (g':guar),
  (forall s, p' s -> p s /\ forall s', g s s' -> g' s s') ->
  iseqOkS P p' g' ii.
Proof.
  intros.
  eapply iseqOkWeak.
  eapply WpSound.
  eauto.
  assumption.
Qed.


(* some handy tactics *)

Ltac iseqWeakSolver' Psolver P :=
  match goal with
    | |- iseqOkSWP _ _ _ _ => econstructor
    | |- P _ = SomeT _ => Psolver
    | |- _ => reflexivity
  end.

Ltac iseqWeakSolver Psolver := 
  match goal with
    | |- iseqOkS ?P _ _ _ => 
      unfold bne; unfold beq;
      eapply UseWeakIseq;
	[repeat (iseqWeakSolver' Psolver P) | idtac]
  end.


Lemma weakSeqOkUnique : forall P p g ii,
  iseqOkSWP P p g ii -> forall p' g',
  iseqOkSWP P p' g' ii ->
  p = p' /\ g = g'.
Proof.
  intros P p g ii iiOk; induction iiOk; subst;
    intros p3 g3; inversion_clear 1; subst; auto.

  rewrite H1 in H; injection H; clear H H1; intros; subst; intuition.

  rewrite H1 in H; injection H; clear H H1; intros; subst.
  destruct (IHiiOk _ _ H2) as [E1 E2]; subst; intuition.

  destruct (IHiiOk _ _ H0) as [E1 E2]; subst; intuition.

  rewrite H3 in H; rewrite H4 in H0; injection H; injection H0; 
    clear H3 H H4 H0; intros; subst; intuition.

Qed.


Fixpoint weakSide (P:chspec) (ii:iseq) {struct ii} : Prop :=
  match ii with
    | jr r => r = ra
    | j f => exists sp, P f = SomeT sp
    | br bt rs rt f ii' => (exists sp, P f = SomeT sp) /\ weakSide P ii'
    | seq c ii' => weakSide P ii'
    | jal f fret => 
      (exists p, exists g, P f = SomeT (bspeci p g) /\ 
	forall s s', p s -> g s s' -> rget s ra = rget s' ra) /\
      (exists sp, P fret = SomeT sp)
  end.


Lemma weakSideSound : forall P ii,
  weakSide P ii ->
  exists p, exists g, iseqOkSWP P p g ii.
Proof.
  induction ii; simpl.

  intros X; destruct (IHii X) as [p [g iiOk]]; clear IHii X.
  econstructor; econstructor; econstructor; eauto.

  intros [[[p' g'] E] IH]; destruct (IHii IH) as [p [g iiOk]]; clear IHii IH.
  econstructor; econstructor; econstructor; eauto.

  intros [[p' g'] E]; econstructor; econstructor; econstructor; eauto.

  intros [[p [g [E X]]] [[p' g'] E']].
  econstructor; econstructor; econstructor; eauto.

  intro; subst r; econstructor; econstructor; econstructor; eauto.

Qed.


Lemma weakSideSound' : forall P p g ii,
  iseqOkSWP P p g ii ->
  weakSide P ii.
Proof.
  induction 1; simpl; subst; auto.
  
  exists (bspeci p' g'); assumption.

  rewrite H; constructor; [econstructor; reflexivity | assumption].

  rewrite H; rewrite H0; clear H H0.
  constructor.
  exists p'; exists g'; intuition.
  econstructor; reflexivity.
  
Qed.



